// $Id: jogshuttle.cc,v 1.10.2.2 2003/01/23 22:01:35 freefrag Exp $
/*
 * Copyright (C) 2001 Tomoaki Hayasaka <hayasakas@postman.riken.go.jp>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
 /// Support for special USB Jog/Shuttle input controllers

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <linux/input.h>

#include "jogshuttle.h"
#include "preferences.h"
#include "commands.h"

#ifdef USE_HID_JOGSHUTTLE
#include <string>
using std::string;
#include <utility>
using std::pair;
using std::make_pair;
#endif

/* **********************************************************************

Overview of jogshuttle mapping code.

Note, this is mostly about the preferences, but I thought it should go in here

* Knowing all the actions

When starting, we fill the optionmenus with the choices. These are
also into an array. Array is keyed on _option_index (duplicated),
contains _short_desc, _desc, _action. When we fill the option menu,
the index is maintained/duplicated in the array.

* Getting the stored or default actions

Then the preference.cc file needs to read a lot of stuff from the
config file or a default mapping. These are ints that are the codes of
the buttons, and the strings that are the actual actions. 

These are read into the mappings of buttons, which is a map keyed on a
pair of codes. The codes are (for now) the actual codes as returned
from the events - a future level of abstraction may call for a mapping
from buttons to codes.

* Changing the mappings - the user changes the optionmenu selections.

- When the user changes one of the two "buttons", we look up the
  action in the JogShuttleMapping map. The action is then located in
  the JogShuttleActions array and the right index is set for the
  Action optionmenu.

- Whenever the action optionmenu is changed, we get the action string
  from the list of all actions (keyed on index) and stuff this into
  the mappings. In other words, changing the action optionmenu,
  "saves" this choice.

* Using the mappings

The jogshuttle.cc code only needs a mapping from code, code to string,
which it gets from the Preference object.

* Saving the mappings

When saving, the stuff the action mapping into the config file.

********************************************************************** */


extern "C" {
#include "support.h"
	extern struct navigate_control g_nav_ctl;
}

/// JogShuttle

/// A C wrappper for the GDK input handler callback
void JogShuttle_inputCallback(gpointer data, gint source, GdkInputCondition condition)
{
  JogShuttle *js = static_cast<JogShuttle *>(data);
  g_return_if_fail(js != NULL);
  js->inputCallback(source, condition);
}

/** The global JogShuttle object.
	 
    It is accesible to other classes via the getInstance method.
*/

JogShuttle *JogShuttle::_instance = NULL;

/** Singleton getInstance method. Standard pattern. Non-threadsafe, but
	first instance gets called from the commands.cc which should be OK.
*/

JogShuttle &JogShuttle::getInstance() {
	if ( _instance == NULL )
		_instance = new JogShuttle();
	return *_instance;
}

/** Constructor

    If enabled in preferences, open the user-specified device and
    hook our callback function into the GDK input handler.
*/
JogShuttle::JogShuttle() :
	_callback(NULL),
	input_(-1),
	monitorTag_(-1)
#ifdef USE_HID_JOGSHUTTLE
					      ,
	last_dial_value(0xFFFFFFFF),
	last_wheel_value(0),
	_modifier_code(0)
#endif
{
	// g_print( "In JogShuttle::JogShuttle\n" );
	Preferences &prefs = Preferences::getInstance();
	if ( prefs.enableJogShuttle ) 
	{
		char *filename = prefs.jogShuttleDevice;
		input_ = open( filename, O_RDONLY );
		if ( input_ < 0 ) {
			// g_print( "JogShuttle::JogShuttle: open %s: %s\n", 
			// 	 filename, g_strerror(errno) );
			return;
		}
		monitorTag_ = gdk_input_add( input_, GDK_INPUT_READ,
					     JogShuttle_inputCallback, 
					     (gpointer)this );
	}
}

/** Destructor

    Remove the callback function from the GDK input handler.
*/
JogShuttle::~JogShuttle()
{
  if (input_ >= 0) {
    gdk_input_remove(monitorTag_);

    close(input_);
    input_ = -1;
  }
}

/** Register callback

    Register an interest in getting notification when buttons are pressed
    
    \param callback Function to call
*/
void JogShuttle::registerCallback(void * user, JogShuttleCallback callback) 
{
	if ( _callback ) 
		g_print( "JogShuttle::registerCallback - already registered\n" );
	_callback     = callback;
	_callbackdata = user;
}

/** Deegister callback

    Deregister an interest in getting notification when buttons are pressed
    
    \param callback Function to stop calling
*/
void JogShuttle::deregisterCallback()
{
	if ( _callback == NULL ) 
		g_print( "JogShuttle::deregisterCallback - not registered\n" );
	_callback = NULL;
}

/** Handle movement on the jog dial.

    \param dir A number from -x to x to specify the offset from the current frame.
               A negative number moves backward. Typically, x is 1 to step frame-by-frame.
*/
void JogShuttle::jog(int dir)
{
	gdk_threads_enter();
	if (dir < 0) videoBack();
	else videoForward();
	gdk_threads_leave();
}

/** Handle movement of the shuttle ring.

    \param angle A number from -8 to 8 that speccifies a direction and speed.
*/
void JogShuttle::shuttle(int angle)
{
	if (angle > 0) angle--;
	if (angle < 0) angle++;
	gdk_threads_enter();
	videoShuttle( angle );
	gdk_threads_leave();
}

/** Handle a button press.
    
    This executes commands directly instead of using the key_press callback.
    
    \param code The keybaord code to which a button is mapped.
    \todo consolidate commands with keyboard input
*/
void JogShuttle::button(int code)
{
  switch (code) {

  case 0xf1:  // Paste before
	gdk_threads_enter();
	processMenuCommand( "P" );
	gdk_threads_leave();
    break;
    
  case 0xf2:  // set in point (d0)
	gdk_threads_enter();
  	processMenuCommand( "d0" );
	gdk_threads_leave();
    break;

  case 0xf3:  // set out point (d$)
	gdk_threads_enter();
  	processMenuCommand( "d$" );
	gdk_threads_leave();
    break;

  case 0xf4:  // paste after
	gdk_threads_enter();
  	processMenuCommand( "p" );
	gdk_threads_leave();
    break;

  case 0xc2:  // jump to start of movie
	gdk_threads_enter();
  	videoStart();
	gdk_threads_leave();
    break;
    
  case 0xc3:  // prev sequence
	gdk_threads_enter();
  	videoPreviousScene();
	gdk_threads_leave();
    break;
    
  case 0xc1:  // play
    if (g_nav_ctl.active == FALSE) {
		gdk_threads_enter();
		videoPlay();
		gdk_threads_leave();
    } else {
		gdk_threads_enter();
        videoPause();
		gdk_threads_leave();
    }
    break;
    
  case 0xc4:  // next sequence
	gdk_threads_enter();
  	videoNextScene();
	gdk_threads_leave();
    break;
    
  case 0xc5:  // jump to end of movie
	gdk_threads_enter();
  	videoEndOfMovie();
	gdk_threads_leave();
    break;

  default:
    break; //do nothing

  }
}

/** The GDK input callback function.

    GDK calls this whenever input is received. It is hooked into the system
    during object construction.
*/
void JogShuttle::inputCallback(gint source, GdkInputCondition condition)
{
  g_return_if_fail(this != NULL);
  g_return_if_fail(input_ >= 0);
  g_return_if_fail(input_ == source);
  g_return_if_fail(condition == GDK_INPUT_READ);

  struct input_event ev;
  ssize_t n = read(input_, &ev, sizeof(ev));
  if (n != sizeof(ev)) {
    g_print("JogShuttle::inputCallback: read: %s\n", g_strerror(errno));
    return;
  }

#if 0
  printf("JogShuttle:");
  for (int i = 0; i < n; i++) {
    printf(" %02x", ((unsigned char *)(&ev))[i]);
  }
  putchar('\n');
#endif

#ifndef USE_HID_JOGSHUTTLE

  if (ev.type != EV_KEY || ev.value == 0)
    return;

  switch (ev.code) {
  case  2: this->jog(-1); break;
  case  3: this->jog( 1); break;
  case 30: this->shuttle(1); break;
  case 48: this->shuttle(2); break;
  case 46: this->shuttle(3); break;
  case 32: this->shuttle(4); break;
  case 18: this->shuttle(5); break;
  case 33: this->shuttle(6); break;
  case 34: this->shuttle(7); break;
  case 35: this->shuttle(8); break;
  case 23: this->shuttle(-1); break;
  case 36: this->shuttle(-2); break;
  case 37: this->shuttle(-3); break;
  case 38: this->shuttle(-4); break;
  case 50: this->shuttle(-5); break;
  case 13: this->shuttle(-6); break;
  case 27: this->shuttle(-7); break;
  case 43: this->shuttle(-8); break;
  
  // DRD> I do not know why Tomoaki chose these input codes.
  case 19: this->button(0xb1); break;
  case 49: this->button(0xf1); break;
  case 24: this->button(0xf2); break;
  case 25: this->button(0xf3); break;
  case 16: this->button(0xf4); break;
  case 21: this->button(0xc1); break;
  case 47: this->button(0xc2); break;
  case 45: this->button(0xc3); break;
  case 31: this->button(0xc4); break;
  case 44: this->button(0xc5); break;
  case 17: this->button(0xc6); break;
  case 20: this->button(0xe1); break;
  case 22: this->button(0xe2); break;
  default:
    g_print("JogShuttle::inputCallback: unknown button code 0x%02x\n", ev.code);
    break;
  };

  return;
#else

  /* We can get three "kinds" of events:

     ev.type == EV_KEY (0x01)
     This is a button press. the 13 Keys are laid out from top to
     button, left to right, from 0x100 to 0x10c
     The actual key that was pressed is in ev.code, ev.value is 0x01 for 
     key pressed, 0x00 for key released.

     ev.type == EV_REL (0x02)
     For this type, there are two "codes":

     - ev.code == REL_WHEEL (0x08) the outer wheel ev.value indicates
       the position, with positive values from 0x01 to 0x07
       (inclusive) indicating clockwise twist, and negative values
       from -1 to -7 indicating counterclockwise twist.  If this value
       is different from 0, apparently it is reported approx. every
       second. Which is annoying. A 0 value, however, is not
       reported. Which is even more annoying.

     - ev.code == REL_DIAL (0x07) the inner dial
       This event is issued when ever the dial position changes or
       approx. every second (or thereabout). Also annoying.
       ev.value takes values from 0x01 to 0xff inclusive and increases
       with clockwise rotation, decreases with counterclockwise rotation. 
       Apparantly a value for 0x00 is _not_ issued - this could be problematic!
       (A "step" on the dial is missed, it seems. I do not think this is a good
       thing). 
     
     Note: Theoretically one could keep an eye on the events, and if
     none are reported for more than a second, it must mean that 0 is the current
     value for the missing event. However, the latency is far to great for
     that to be of any use.
     I have no idea if these problems comes from gnome/gtk/gdk or is a real 
     kernel issue.
     The consequences are that the outer ring only have 7 working values and
     that the inner ring will "miss a step", once in a while. Annoying.
     
     Modifier keys - we maintain the state of the first key that is
     pressed down. This is a "poor man"s way of doing modifier
     keys. It means that we can press a single key, keep it down, and
     press a number of others. Any key can be used as a modifier key,
     although you would probably want to not assign an action to a
     modifier key when first pressed. There is no way to react to multiple
     modifier keys though.

   */
  
  /* Check the relative motions */
  if (ev.type == EV_REL) {
    /* First check the outer dial */
    if (ev.code == REL_WHEEL) {
      // g_print("REL_WHEEL = 0x%02x\n", ev.value);
      /* We only want to shuttle if needed - otherwise kino
	 will stop and restart the video thread */
      if (ev.value != last_wheel_value) {
	this->shuttle(ev.value);
	last_wheel_value = ev.value;
      }
      return;
    }
    
    /* Handle the inner dial */
    if (ev.code == REL_DIAL) {
      // g_print("JogShuttle::inputCallback: REL_DIAL, value=0x%02x\n",
      // ev.value);
      if (0xFFFFFFFF == last_dial_value) {
	last_dial_value = ev.value;
	return;
      }
      /* Handling passing 0x00 
	 I think this interval (20) is highly pessimistic. */
      if (ev.value < 10 && last_dial_value > 0xF0) {
	/* We probably crossed clockwise */
	this->jog(1);
	last_dial_value = ev.value;
	return;
      }
      if (ev.value > 0xF0 && last_dial_value < 0x10) {
	/* We probably crossed counter clockwise */
	this->jog(-1);
	last_dial_value = ev.value;
	return;
      }
      if (ev.value > last_dial_value) {
	this->jog(1);
	last_dial_value = ev.value;
	return;
      } else {
	if (ev.value < last_dial_value) {
	  this->jog(-1);
	  last_dial_value = ev.value;
	  return;
	}
      }
      /* Same value, return */
      return;
    }
  }

  /* Now the button _presses_ 
     Release may need to clear a modifier key 
  */
  if ( ev.type != EV_KEY )
	  return;
  if ( ev.value == 0 ) 
  {
	  if ( ev.code == _modifier_code ) 
	  {
		  // g_print( "JogShuttle::inputCallback, modifier 0x%02x cleared\n", 
		  //	ev.code );
		  _modifier_code = 0;
	  }
	  return;
  }
  /* This is a key press - if there are no modifier, make sure that
     this is saved */
  if ( _modifier_code == 0 )
  {
	  _modifier_code = ev.code;
	  // g_print( "JogShuttle::inputCallback, modifier button code 0x%02x\n", 
	  //   ev.code );
  }
  
  // g_print( "JogShuttle::inputCallback: button code 0x%02x\n", ev.code );
  /* Figure out what codes to use */
  unsigned short first;
  unsigned short second;
  if ( _modifier_code != ev.code ) 
  {
	  first  = _modifier_code;
	  second = ev.code;
  }
  else
  {
	  first  = ev.code; /* Same as modifier */
	  second = 0;
  }
  // g_print( "JogShuttle::inputCallback: checking map for 0x%02x, 0x%02x\n", 
  // first, second );
  
  /* Do a callback or action based command */
  if ( _callback != NULL ) 
  {
	  _callback( _callbackdata, first, second );
  }
  else
  {
	  /* Get the action from the prefs */
	  string action
		  = Preferences::getInstance()._JogShuttleMappings[ 
			  make_pair( first, second ) ]._action;
	  
	  if ( "" != action ) 
	  {
		  // g_print( "Processing %s\n", action.c_str() );
		  gdk_threads_enter();
		  processMenuCommand( (char *) action.c_str() );
		  gdk_threads_leave();
	  } 
	  else 
	  {
		  // g_print( "No action\n" );
	  }
  }
	  return;

#endif  
}
