/*
 * page_capture.cc Notebook Firewire Capture Page Object
 * Copyright (C) 2001 Charles Yates <charles.yates@pandora.be>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
using std::cout;
using std::endl;

#include <gnome.h>
#include <math.h>
#include <limits.h>

#include "page_capture.h"
#include "riff.h"
#include "avi.h"
#include "playlist.h"
#include "filehandler.h"
#include "ieee1394io.h"
#include "framedisplayer.h"
#include "message.h"
#include "error.h"
#include "frame.h"

static FrameDisplayer *displayer = NULL;

extern "C" {
	#include "support.h"
	#include "callbacks.h"
	#include "interface.h"
	#include "commands.h"

	#include <pthread.h>

	extern struct navigate_control g_nav_ctl;

    // reader changed to pointer to delay creation of Frame
    // objects so prefs dv decoder quality is properly read
    IEEE1394Reader *reader = NULL;
	AVC *avc = NULL;
    static FileHandler *writer = NULL;
	extern char cmd[];
	static char lastcmd[ 256 ];
	static gchar lastPreferenceFilename[ 512 ];
	extern KinoCommon *common;
	static quadlet_t avcStatus;
    char timecode[ 256 ];

	//
	// Preview rendering implementation
	//
	
	static void *captureThread( void *p );
	static void *avcThread( void *p );
    static void *videoThread(void *p);
	static pthread_t capturethread;
	static pthread_t avcthread;
	static pthread_t videothread;
	static gboolean audioOn = FALSE;
	static pthread_mutex_t writerlock = PTHREAD_MUTEX_INITIALIZER;

	// Define the size of the buffer
	#define BUFFERED_FRAMES		25

	// Buffer and positional information
	static int framePosition = -1;
	static Frame *frameBuffer[ BUFFERED_FRAMES ];
	static int frameCount = 0;
	static void TriggerAction( );
}

/** Constructor for page.

  	\param common	common object to which this page belongs
*/

PageCapture::PageCapture( KinoCommon *common ) : isCapturing(false), captureMutex(false),
	saveAVCActive( false ), avc_enabled( false ), driver_locked( false )
{
	cout << "> Creating Capture Page" << endl;
	this->common = common;
	this->frameArea = GTK_DRAWING_AREA( lookup_widget( common->getWidget(), "capture_drawingarea" ) );
	this->recordButton = GTK_TOGGLE_BUTTON( lookup_widget( common->getWidget(), "capture_page_record_button" ) );
	this->stopButton = GTK_TOGGLE_BUTTON( lookup_widget( common->getWidget(), "capture_page_stop_button" ) );
	this->muteButton = GTK_TOGGLE_BUTTON( lookup_widget( common->getWidget(), "capture_page_mute_button" ) );
	this->fileEntry = GTK_ENTRY( lookup_widget(common->getWidget(), "capture_page_file_entry") );
	this->timecodeLabel = GTK_LABEL( lookup_widget( common->getWidget(), "position_label" ) );
	lastPreferenceFilename[0] = 0;
	memset( timecode, 0, sizeof( timecode ) );
}

/** Destructor for page.
*/

PageCapture::~PageCapture() {
	cout << "> Destroying Capture Page" << endl;
	delete reader;
	reader = NULL;
}

/** Called when a new file operation is selected by the user.
*/

void PageCapture::newFile( )
{
	strcpy( lastPreferenceFilename, "" );
}

/** Start of page.
*/

void PageCapture::start() {
	cout << ">> Starting Capture" << endl;

	// By default, AVC is always disabled
	driver_available = false;
	check_phyid = true;
	avc_enabled = false;
	validComponents = 0;
	gui_state_was = -1;

	// Create the reader object
	if ( Preferences::getInstance().dvCaptureDriver == 0)
		reader = new raw1394Reader( Preferences::getInstance().interface, 
								 Preferences::getInstance().channel, 
								 Preferences::getInstance().dvCaptureBuffers );
	else
		reader = new dv1394Reader( Preferences::getInstance().channel, 
								 Preferences::getInstance().dvCaptureBuffers );
	
	// Create the AV/C object
	avc = new AVC( Preferences::getInstance().interface );

	// Determine the capture directory and file - rules are:
	// 1) if a full path is specified in the preferences, then use that
	// 2) if a non-full path is specified, then pick up the project directory and attach preference
	
	string capture_file_name = Preferences::getInstance().file;
	if ( capture_file_name[ 0 ] != '/' )
		capture_file_name = common->getPlayList( )->GetProjectDirectory( ) + "/" + capture_file_name;

	// If the preferences have changed, then update the filename
	
	if ( capture_file_name != lastPreferenceFilename )
	{
		strcpy( lastPreferenceFilename, capture_file_name.c_str( ) );
		gtk_entry_set_text( fileEntry, capture_file_name.c_str( ) );
	}

	// Have the GUI reflect the mute status
	gtk_toggle_button_set_active( this->muteButton, !audioOn );

	// Create the FrameDisplayer
	displayer = new FrameDisplayer;

	// Set up the preview rendering
	pthread_create( &capturethread, NULL, captureThread, this );
	pthread_create( &avcthread, NULL, avcThread, this );
	pthread_create( &videothread, NULL, videoThread, lookup_widget( common->getWidget(), "capture_drawingarea") );

	common->activateWidgets();
	gtk_label_set_text( timecodeLabel, "");
	common->packIt( "packer_capture" );
	common->toggleComponents( VIDEO_STOP, false);
	common->commitComponentState();
}

/** Check devices and status.
*/

void PageCapture::CheckDevices( )
{
	static int avcRetry = 0;
	
	if ( driver_locked == true )
	{
		cout << "Stopping thread" << endl;
		reader->StopThread( );
		driver_locked = false;
		driver_available = false;
		avc_enabled = false;
		check_phyid = true;
		TriggerAction( );
	}
	else if ( driver_available == false )
	{
    	// Start the reader thread 
		//cout << "Starting thread" << endl;
    	if ( reader->StartThread() == true ) 
		{
			// The drivers are installed
			driver_available = true;
			avc_enabled = false;

        	// Check if the id of the device is valid
			Preferences::getInstance( ).phyID = avc->isPhyIDValid( avc->getNodeId(Preferences::getInstance( ).avcGUID) );
			Preferences::getInstance( ).phyID = avc->getNodeId(Preferences::getInstance( ).avcGUID);
			check_phyid = Preferences::getInstance( ).phyID < 0;
    	}
		TriggerAction( );
	}
	else if ( check_phyid )
	{
		// Check that we're still valid.
		Preferences::getInstance( ).phyID = avc->isPhyIDValid( avc->getNodeId(Preferences::getInstance( ).avcGUID) );
		Preferences::getInstance( ).phyID = avc->getNodeId(Preferences::getInstance( ).avcGUID);
		check_phyid = Preferences::getInstance( ).phyID < 0;
		avc_enabled = false;
		TriggerAction( );
	}
	else if ( avc_enabled )
	{
		quadlet_t status = avc->TransportStatus( Preferences::getInstance().phyID );

		if ( (int) status < 0 )
		{
			cerr << ">>> AVC status error" << endl;
			if ( ++avcRetry > 2 )
			{
				avcRetry = 0;
				avc_enabled = false;
				check_phyid = true;
			}
		}
		else
		{
           	char t[12];

			avcStatus = status;
			avcRetry = 0;
			
			if ( avc->Timecode( Preferences::getInstance().phyID, t ) )
			{
				strncpy( timecode, t, 12 );
			}
			TriggerAction();
		}
	}
	else 
	{
		avc->Noop();
	}
}

/** Reflect the the AV/C status in the transport buttons.
*/
void PageCapture::applyAVCState( quadlet_t newAVCState )
{
	if ( driver_available == false  )
	{
		if ( gui_state_was != 0 )
		{
			gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_options_button" ), false );
			gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_record_button" ), false );
			gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_stop_button" ), false );
			gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_snapshot_button" ), false );
			gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_mute_button" ), false );
			validComponents = 0;
			common->activateWidgets( );
    		common->commitComponentState();
			common->setStatusBar( "Drivers not loaded?" );
		}
		gui_state_was = 0;
	}
	else if ( check_phyid == true )
	{
		if ( gui_state_was != 1 )
		{
			gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_options_button" ), false );
			gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_record_button" ), true );
			gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_stop_button" ), true );
			gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_snapshot_button" ), true );
			gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_mute_button" ), true );
			validComponents = 0;
			common->activateWidgets( );
    		common->commitComponentState();
			common->setStatusBar( "No AV/C compliant cam connected or not switched on?");
		}
		gui_state_was = 1;
	}
	else if ( avc_enabled )
	{
		if ( gui_state_was != 2 )
		{
			gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_options_button" ), true );
			gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_record_button" ), true );
			gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_stop_button" ), true );
			gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_snapshot_button" ), true );
			gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_mute_button" ), true );

			validComponents = VIDEO_REWIND |
				VIDEO_PLAY |
				VIDEO_PAUSE |
				VIDEO_STOP |
				VIDEO_FAST_FORWARD |
				VIDEO_FORWARD |
				VIDEO_BACK |
				VIDEO_NEXT_SCENE |
				VIDEO_START_OF_SCENE |
				VIDEO_START_OF_MOVIE |
				VIDEO_SHUTTLE |
				VIDEO_END_OF_MOVIE | 
				INFO_FRAME;

			common->activateWidgets( );

			gui_state_was = 2;

			// common->setStatusBar( "AV/C Controls Enabled" );
			cerr << ">>> AVC enabled " << endl;

			// Ensure that we reflect the current avc state when changing gui states
			avcState = !avcState;
		}

		if ( avcState != newAVCState )
		{
    		quadlet_t resp2 = AVC1394_MASK_RESPONSE_OPERAND(newAVCState, 2);
    		quadlet_t resp3 = AVC1394_MASK_RESPONSE_OPERAND(newAVCState, 3);
    		this->avcState = newAVCState;
    
    		common->toggleComponents( common->getComponentState(), false );
			g_nav_ctl.active = FALSE;

    		if (resp2 == AVC1394_VCR_RESPONSE_TRANSPORT_STATE_PLAY) {
        		if (resp3 >= AVC1394_VCR_OPERAND_PLAY_FAST_FORWARD_1
                		&& resp3 <= AVC1394_VCR_OPERAND_PLAY_FASTEST_FORWARD) {
            		common->toggleComponents( VIDEO_FAST_FORWARD, true );
					g_nav_ctl.active = TRUE;
        		} else if (resp3 >= AVC1394_VCR_OPERAND_PLAY_FAST_REVERSE_1
                   		&& resp3 <= AVC1394_VCR_OPERAND_PLAY_FASTEST_REVERSE) {
            		common->toggleComponents( VIDEO_REWIND, true );
					g_nav_ctl.active = TRUE;
        		} else if (resp3 == AVC1394_VCR_OPERAND_PLAY_FORWARD_PAUSE) {
            		common->toggleComponents( VIDEO_PAUSE, true );
        		} else {
            		common->toggleComponents( VIDEO_PLAY, true );
					g_nav_ctl.active = TRUE;
        		}
    		} else if (resp2 == AVC1394_VCR_RESPONSE_TRANSPORT_STATE_WIND) {
        		if (resp3 == AVC1394_VCR_OPERAND_WIND_HIGH_SPEED_REWIND) {
            		common->toggleComponents( VIDEO_START_OF_MOVIE, true );
        		} else if (resp3 == AVC1394_VCR_OPERAND_WIND_STOP) {
            		common->toggleComponents( VIDEO_STOP, true );
        		} else if (resp3 == AVC1394_VCR_OPERAND_WIND_REWIND) {
            		common->toggleComponents( VIDEO_START_OF_MOVIE, true );
        		} else if (resp3 == AVC1394_VCR_OPERAND_WIND_FAST_FORWARD) {
            		common->toggleComponents( VIDEO_END_OF_MOVIE, true );
        		} else {
            		cout << "AVC Status: Unkown winding" << endl;
        		}
    		} else {
        		cout << "AVC Status: Unknown state" << endl;
    		}
    		common->commitComponentState();
		}
	}
	else
	{
		if ( gui_state_was != 3 )
		{
			gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_options_button" ), true );
			gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_record_button" ), true );
			gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_stop_button" ), true );
			gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_snapshot_button" ), true );
			gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_mute_button" ), true );
			validComponents = 0;
			common->activateWidgets( );
    		common->commitComponentState();
			if ( !isCapturing )
				common->setStatusBar( "AV/C Controls Not Enabled." );
			gui_state_was = 3;
		}
	}
}

/** Define active widgets.
*/

gulong PageCapture::activate() {
	return this->validComponents;
}

/** Leaving the page
*/

void PageCapture::clean() {
	cout << ">> Leaving Capture" << endl;
	common->setStatusBar( "" );
	if ( g_nav_ctl.capture_active == TRUE ) {

		// Stop the capture
		stopCapture();

		// Stop the video
		if (Preferences::getInstance( ).phyID >= 0)
			videoStop();
		
		// Start the deactivation process for the backround processes
		g_nav_ctl.active = FALSE;
		g_nav_ctl.capture_active = FALSE;
		reader->TriggerAction( );

		// Stop the video thread
		gdk_threads_leave();
		pthread_join( videothread, NULL );
		gdk_threads_enter();

		// Join with the capture thread to make sure we're definitely finished
		pthread_join( capturethread, NULL );
		pthread_join( avcthread, NULL );

		// Stop the reading thread
   		reader->StopThread();
	}

	// Clean up objects created in start
   	delete displayer;
	displayer = NULL;
   	delete reader;
	reader = NULL;
	delete avc;
	avc = NULL;
}


/** Start of movie.
*/

void PageCapture::videoStartOfMovie() {
	if ( avc_enabled ) {
		audioOn = FALSE;
		gtk_toggle_button_set_active( this->muteButton, !audioOn );
		common->toggleComponents( VIDEO_START_OF_MOVIE, avc_enabled);
    	if (avc->Stop( Preferences::getInstance().phyID) >= 0)
    		avc->Rewind( Preferences::getInstance().phyID);
	}
}

/** Start of previous scene.
*/

void PageCapture::videoPreviousScene() {
	if ( avc_enabled ) {
		audioOn = FALSE;
		gtk_toggle_button_set_active( this->muteButton, !audioOn );
		common->toggleComponents( VIDEO_START_OF_SCENE, avc_enabled);
    	if (avc->Pause( Preferences::getInstance().phyID ) >= 0)
    		avc->PreviousScene( Preferences::getInstance().phyID);
	}
}

/** Start of current scene.
*/

void PageCapture::videoStartOfScene() {
	if ( avc_enabled ) {
		audioOn = FALSE;
		gtk_toggle_button_set_active( this->muteButton, !audioOn );
		common->toggleComponents( VIDEO_START_OF_SCENE, avc_enabled);
    	if (avc->Pause( Preferences::getInstance().phyID ) >= 0)
    		avc->PreviousScene( Preferences::getInstance().phyID);
	}
}

/** Rewind.
*/

void PageCapture::videoRewind() {
	if ( avc_enabled ) {
		audioOn = FALSE;
		gtk_toggle_button_set_active( this->muteButton, !audioOn );
		common->toggleComponents( VIDEO_REWIND, avc_enabled);
    	if (avc->Pause( Preferences::getInstance().phyID ) >= 0)
    		avc->Rewind( Preferences::getInstance().phyID);
	}
}

/** Frame back.
*/

void PageCapture::videoBack() {
	if ( avc_enabled ) {
		audioOn = FALSE;
		gtk_toggle_button_set_active( this->muteButton, !audioOn );
		common->toggleComponents( VIDEO_BACK, avc_enabled);
    	if (avc->Pause( Preferences::getInstance().phyID ) >= 0)
    		avc->Back( Preferences::getInstance().phyID);
	}
}

/** Play.
*/

void PageCapture::videoPlay() {
	if ( avc_enabled ) {
		audioOn = TRUE;
		gtk_toggle_button_set_active( this->muteButton, !audioOn );
		common->toggleComponents( VIDEO_PLAY, avc_enabled);
    	avc->Play( Preferences::getInstance().phyID);
	}
}

/** Pause.
*/

void PageCapture::videoPause() {
	if ( avc_enabled ) {
		audioOn = FALSE;
		gtk_toggle_button_set_active( this->muteButton, !audioOn );
		common->toggleComponents( VIDEO_PAUSE, avc_enabled);
    	avc->Pause( Preferences::getInstance().phyID);
	}
}

/** Stop.
*/

void PageCapture::videoStop() {
	if ( avc_enabled ) {
		audioOn = FALSE;
		gtk_toggle_button_set_active( this->muteButton, !audioOn );
		common->toggleComponents( VIDEO_STOP, avc_enabled);
    	avc->Stop( Preferences::getInstance().phyID);
	}
}

/** Forward.
*/

void PageCapture::videoForward() {
	if ( avc_enabled ) {
		audioOn = FALSE;
		gtk_toggle_button_set_active( this->muteButton, !audioOn );
		common->toggleComponents( VIDEO_FORWARD, avc_enabled);
    	if (avc->Pause( Preferences::getInstance().phyID ) >= 0)
    		avc->Forward( Preferences::getInstance().phyID);
	}
}

/** Fast Forward.
*/

void PageCapture::videoFastForward() {
	if ( avc_enabled ) {
		audioOn = FALSE;
		gtk_toggle_button_set_active( this->muteButton, !audioOn );
		common->toggleComponents( VIDEO_FAST_FORWARD, avc_enabled);
    	if (avc->Pause( Preferences::getInstance().phyID ) >= 0)
    		avc->FastForward( Preferences::getInstance().phyID);
	}
}

/** Start of next scene.
*/

void PageCapture::videoNextScene() {
	if ( avc_enabled ) {
		audioOn = FALSE;
		gtk_toggle_button_set_active( this->muteButton, !audioOn );
		common->toggleComponents( VIDEO_NEXT_SCENE, avc_enabled);
    	if (avc->Pause( Preferences::getInstance().phyID ) >= 0)
    		avc->NextScene( Preferences::getInstance().phyID);
	}
}

/** End of current scene.
*/

void PageCapture::videoEndOfScene() {
	if ( avc_enabled ) {
		audioOn = FALSE;
		gtk_toggle_button_set_active( this->muteButton, !audioOn );
		common->toggleComponents( VIDEO_NEXT_SCENE, avc_enabled);
    	if (avc->Pause( Preferences::getInstance().phyID ) >= 0)
    		avc->NextScene( Preferences::getInstance().phyID);
	}
}

/** End of movie.
*/

void PageCapture::videoEndOfMovie() {
	if ( avc_enabled ) {
		audioOn = FALSE;
		gtk_toggle_button_set_active( this->muteButton, !audioOn );
		common->toggleComponents( VIDEO_END_OF_MOVIE, avc_enabled);
    	if (avc->Stop( Preferences::getInstance().phyID) >= 0)
    		avc->FastForward( Preferences::getInstance().phyID);
	}
}

/** Shuttle.

	Bi-directional variable-speed playback.
	
	\param angle The relative speed.
*/

void PageCapture::videoShuttle( int angle ) {
	if ( avc_enabled ) {
		audioOn = FALSE;
    	if (angle != 0)
	    	avc->Shuttle( Preferences::getInstance().phyID, angle*2 );
    	else
	    	avc->Pause( Preferences::getInstance().phyID );
	}
}


/** Process a keyboard event. 

  	\param event	keyboard event
*/

gboolean PageCapture::processKeyboard( GdkEventKey *event ) {
	gboolean ret = FALSE;

	// Only process while not escape mode
	if ( g_nav_ctl.escaped == FALSE ) {
		// Translate special keys to equivalent command
		switch (event->keyval) {
			case GDK_BackSpace:
			case GDK_Left:
				strcat(cmd, "h");
				break;
			case GDK_Up:
				strcat(cmd, "k");
				break;
			case GDK_Right:
				strcat(cmd, "l");
				break;
			case GDK_Return:
				startCapture();
				cmd[0] = 0;
				break;
			case GDK_Down:
				strcat(cmd, "j");
				break;
			case GDK_Escape:
				if (isCapturing) {
					stopCapture();
				} else if (common->getComponentState() & VIDEO_STOP) {
					common->changePageRequest( PAGE_EDITOR );
					return FALSE;
				} else if (avc_enabled && Preferences::getInstance( ).phyID >= 0) {
					common->keyboardFeedback(cmd, "Stop");
					common->videoStop( );
				} else {
					common->changePageRequest( PAGE_EDITOR );
					return FALSE;
				}
				cmd[0] = 0;
				break;
			case GDK_Alt_L:
			case GDK_Alt_R:
				strcpy( lastcmd, "alt");
				return FALSE;
			default:
				strcat(cmd, event->string);
				break;
		}
	
		if ( !strcmp( cmd, "." ) )
			strcpy( cmd, lastcmd );

		ret = processKeyboardCommand( cmd );

	} else if (event->keyval == GDK_Return || event->keyval == GDK_Escape) {
		g_nav_ctl.escaped = FALSE;
		gtk_entry_set_editable( fileEntry, FALSE );
		gtk_widget_grab_focus( common->getWidget() );
	}
	return ret;
}

/** Process a menu command.

  	\param command	command to be processed
*/

gboolean PageCapture::processMenuCommand( char *command ) {
	strcpy( cmd, command );
	gboolean ret = processKeyboardCommand( cmd );
	return ret;
}

/** Internal method for handling a complete keyboard scene.

  	\param cmd		command to be processed;
*/

gboolean PageCapture::processKeyboardCommand( char *cmd ) {
	int count = 1;
	char real[ 256 ] = "";

	if (validComponents == 0) return FALSE;

	switch( sscanf( cmd, "%d%s", &count, real ) )
	{
		case 1:
			// Numeric value only - return immediately if the cmd is not "0"
			if ( strcmp( cmd, "0" ) ) {
				common->keyboardFeedback(cmd, "");
				return FALSE;
			}
			break;
		case 0:
			sscanf( cmd, "%s", real );
			count = 1;
			break;
	}

	if ( strcmp( lastcmd, "alt") == 0 && strcmp( cmd, "f" ) == 0 ) 
	{
		gtk_entry_set_editable( fileEntry, TRUE );
		gtk_widget_grab_focus( GTK_WIDGET(fileEntry) );
		g_nav_ctl.escaped = TRUE;
		cmd[0] = 0;
		return FALSE;
	}

	if (Preferences::getInstance( ).phyID < 0) return FALSE;

	strcpy( lastcmd, cmd );

	/* Navigation */

	/* play, pause */
	
	if ( strcmp(cmd, " ") == 0 ) 
	{
		if ( g_nav_ctl.active == FALSE ) {
			common->keyboardFeedback(cmd, "Play");
			common->videoPlay( );
		} else {
			common->keyboardFeedback(cmd, "Pause");
			common->videoPause( );
		}
		cmd[0] = 0;
	}

	/* advance one frame */

	else if ( strcmp(real, "l") == 0 ) 
	{
		common->keyboardFeedback(cmd, "Move forward" );
		common->videoForward();
		cmd[0] = 0;
	}

	/* backspace one frame */

	else if ( strcmp(real, "h") == 0 ) 
	{
   		common->keyboardFeedback(cmd, "Move backward");
		common->videoBack();
		cmd[0] = 0;
	}

	/* advance one second */

	else if ( (strcmp(real, "w") == 0 || strcmp(real, "W") == 0 ||
		  	strcmp(real, "e") == 0 || strcmp(real, "E") == 0) ) 
	{
		common->keyboardFeedback(cmd, "Fast Forward");
		common->videoFastForward();
		cmd[0] = 0;
	}

	/* backspace one second */

	else if ( (strcmp(real, "b") == 0 || strcmp(real, "B") == 0) ) 
	{
		common->keyboardFeedback(cmd, "Rewind");
		common->videoRewind();
		cmd[0] = 0;
	}

	/* start of scene */

	else if ( (strcmp(cmd, "0") == 0 || strcmp(real, "^") == 0) ) 
	{
		common->videoStartOfScene( );
		for ( ; count > 1; count -- ) 
		{
			common->g_currentFrame --;
			videoStartOfScene( );
		}
		common->keyboardFeedback(cmd, "Move to start of scene");
		cmd[0] = 0;
	}

	/* end of scene */

	else if ( strcmp(real, "$") == 0 ) 
	{
		common->videoEndOfScene( );
		for ( ; count > 1; count -- ) 
		{
			common->g_currentFrame ++;
			videoEndOfScene( );
		}
		common->keyboardFeedback(cmd, "Move to end of scene");
		cmd[0] = 0;
	}

	/* start of next scene */

	else if ( (strcmp(real, "j") == 0 || strcmp(real, "+") == 0) ) 
	{
		for ( ; count >= 1; count -- )
			common->videoNextScene( );
		common->keyboardFeedback(cmd, "Move to start of next scene");
		cmd[0] = 0;
	}

	/* start of previous scene */

	else if ( (strcmp(real, "k") == 0 || strcmp(real, "-") == 0) ) 
	{
		for ( ; count >= 1; count -- )
			common->videoPreviousScene( );
		common->keyboardFeedback(cmd, "Move to start of previous scene");
		cmd[0] = 0;
	}

	/* first frame */

	else if ( strcmp(cmd, "gg") == 0 ) 
	{
		common->videoStartOfMovie( );
		common->keyboardFeedback(cmd, "Move to first frame");
		cmd[0] = 0;
	}

	/* last frame */

	else if ( strcmp(cmd, "G") == 0 ) 
	{
		common->videoEndOfMovie( );
		common->keyboardFeedback(cmd, "Move to last frame");
		cmd[0] = 0;
	}

	else 
	{
		// Check for invalid commands
		if ( strlen( real ) > 3 )
			cmd[ 0 ] = 0;
		else if ( strchr( "dgy ", real[ strlen( real ) - 1 ] ) == NULL ) 
			cmd[ 0 ] = 0;

		common->keyboardFeedback(cmd, "");

#if 0
		printf("send_event: %2.2x\n", event->send_event);
		printf("time  : %8.8x\n", event->time);
		printf("state : %8.8x\n", event->state);
		printf("keyval: %8.8x\n", event->keyval);
		printf("length: %8.8x\n", event->length);
		printf("string: %s\n", event->string);
		printf("(hex) : %2.2x\n", event->string[0]);
		printf("cmd   : %s\n", cmd);
#endif
	}

	return FALSE;
}

/** Start capturing video to the disk.

    If AV/C is enabled, tell the device to play and temporarily 
    disable AV/C and shutdown the AV/C status/timecode monitor.
*/
void PageCapture::startCapture()
{
	if ( !captureMutex && !isCapturing )
	{
		captureMutex = true;
		isCapturing = true;
		gtk_toggle_button_set_active( recordButton, true );
		gtk_toggle_button_set_active( stopButton, false );
		gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_snapshot_button" ), false );
		gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_options_button" ), false );
		captureMutex = false;
		
		char filename[ 512 ];
		strcpy( filename, gtk_entry_get_text( fileEntry ) );

        if ( !strcmp( filename, "" ) )
		{
			common->setStatusBar( "You must enter a filename to capture." );
			captureMutex = true;
			gtk_toggle_button_set_active( recordButton, false );
			gtk_toggle_button_set_active( stopButton, true );
			gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_snapshot_button" ), true );
			gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_options_button" ), true );
			captureMutex = false;
			isCapturing = false;

        }
		else
		{
            Frame *frame = NULL;
			int retry;
			const int MAX_RETRIES = 10;
			struct timespec t = { 1,  0UL };

			gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_options_button" ), false );

			saveAVCActive = avc_enabled;
        	if ( saveAVCActive ) {
				videoPlay( );
				avc_enabled = false;
			}

			FileHandler *handler = NULL;

			switch ( Preferences::getInstance().fileFormat)
			{
				case RAW_FORMAT:
					handler = new RawHandler();
					break;
#ifdef HAVE_LIBQUICKTIME
				case QT_FORMAT:
					handler = new QtHandler();
					break;
#endif
				case AVI_DV1_FORMAT:
				case AVI_DV2_FORMAT:
					AVIHandler *aviWriter = new AVIHandler( Preferences::getInstance().fileFormat);
					handler = aviWriter;
					aviWriter->SetOpenDML( Preferences::getInstance().isOpenDML ||
                        Preferences::getInstance().maxFileSize == 0 ||
                        Preferences::getInstance().maxFileSize > 1024 );
					break;
			}

            /* set command line parameters to the handler object */

            handler->SetMaxFrameCount( Preferences::getInstance().frames);
            handler->SetAutoSplit( Preferences::getInstance().autoSplit);
            handler->SetTimeStamp( Preferences::getInstance().timeStamp);
            handler->SetEveryNthFrame( Preferences::getInstance().every);
			handler->SetMaxFileSize( (off_t)Preferences::getInstance().maxFileSize * (off_t)(1024*1024));
            handler->SetBaseName(filename);

			// without this, if no video has yet played, then frameBuffer is
			// likely to be empty due to latency of avcPlay above
			for ( retry = MAX_RETRIES; retry > 0 && isCapturing; --retry )
			{
				frame = frameBuffer[ framePosition ];
				
				if (frame != NULL) {
					pthread_mutex_lock( &writerlock );
	
					writer = handler;
	
					writer->SetSampleFrame(*frame);
					if ( writer->WriteFrame(*frame) == false ) {
						delete writer;
						writer = NULL;
						pthread_mutex_unlock( &writerlock );
						captureMutex = false;
						stopCapture();
						modal_message( "Failed to write to file. Is file name OK?" );
					}
					else
					{
						common->setStatusBar( "Now capturing..." );
						if (!Preferences::getInstance().preview_capture) windowChanged();
					}
					pthread_mutex_unlock( &writerlock );
					break;
				}						
				common->setStatusBar( "Waiting for DV %s%d ...", 
					( saveAVCActive? "" : "(press play on camera) " ), retry );
				while ( gtk_events_pending() )
					gtk_main_iteration();
				nanosleep( &t, NULL );
			}
			if ( retry == 0 && isCapturing )
			{
				// timed out waiting for iso stream, stop capture
				this->stopCapture();
				common->setStatusBar( "Capture failed." );
			}
			else if ( !isCapturing )
				common->setStatusBar( "Capture stopped." );
        }
    }
	else if ( !captureMutex )
	{
		captureMutex = true;
		gtk_toggle_button_set_active( recordButton, true );
		captureMutex = false;
	}
}

/** Stop capturing video to the disk.

    If AV/C was previously enabled, then re-enable it, pause the
    device and restart the AV/C status/timecode monitor.
*/
void PageCapture::stopCapture() {

	if ( !captureMutex  && isCapturing ) {
		captureMutex = true;

		pthread_mutex_lock( &writerlock );
		if ( writer != NULL )
		{
			writer->Close();
        	delete writer;
        	writer = NULL;
		}
		pthread_mutex_unlock( &writerlock );
        
		gtk_toggle_button_set_active( recordButton, false );
		gtk_toggle_button_set_active( stopButton, true );
		gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_snapshot_button" ), true );
		gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_options_button" ), true );

		audioOn = FALSE;
		gtk_toggle_button_set_active( this->muteButton, !audioOn );

		gtk_widget_set_sensitive( lookup_widget( common->getWidget(), "capture_page_options_button" ), true );

		if ( saveAVCActive ) {
			avc_enabled = saveAVCActive;
			videoPause( );
		}
		common->setStatusBar( "" );
        isCapturing = false;
        captureMutex = false;
    } else if ( !captureMutex ) {
		captureMutex = true;
		gtk_toggle_button_set_active( stopButton, true );
		captureMutex = false;
	}
}


/** Save a single still frame of video to the disk.
*/
void PageCapture::saveFrame() {
	Frame *frame = frameBuffer[ framePosition ];

	if ( frame != NULL ) {
	    unsigned char   pixels[FRAME_MAX_WIDTH*FRAME_MAX_HEIGHT*4];
	  	frame->ExtractRGB( pixels );
	   	GdkImlibImage *im = gdk_imlib_create_image_from_data (pixels, NULL, frame->GetWidth(), frame->GetHeight());
   		char *filename = common->getFileToSave( "Save Still Frame" );
		if ( strcmp( filename, "" ) ) 
		   	gdk_imlib_save_image(im, filename, NULL);
		gdk_imlib_destroy_image( im );
	}
}

void PageCapture::windowChanged() {
	if (frameBuffer[ framePosition ] == NULL)
		common->clearPreview( frameArea );
	else 
		displayer->Put(*frameBuffer[ framePosition ], GTK_WIDGET(frameArea), TRUE); 
}

void PageCapture::windowMoved() {
	if ( g_nav_ctl.capture_active == FALSE )
		common->clearPreview( frameArea );
}


/** Show the AV/C timecode.
*/
void PageCapture::showFrameInfo( int i ) {
	if ( (!isCapturing || Preferences::getInstance().preview_capture) && avc_enabled ) {
		char msg[256];
		sprintf(msg, "-\n-\n%s", timecode );
		gtk_label_set_text( timecodeLabel, msg);
	}
}


/*
 * Callbacks
 */

extern "C"
{
	gchar entry_text[1024];
	
	//
	// Threaded implementation
	//

	static pthread_mutex_t condition_mutex = PTHREAD_MUTEX_INITIALIZER;
	static pthread_cond_t condition = PTHREAD_COND_INITIALIZER;
	
	static int WaitForAction( int lastPosition )
	{
		if ( framePosition == lastPosition )
		{
			pthread_mutex_lock( &condition_mutex );
			pthread_cond_wait( &condition, &condition_mutex );
			pthread_mutex_unlock( &condition_mutex );
		}

		return framePosition;
	}

	static void TriggerAction( )
	{
		pthread_mutex_lock( &condition_mutex );
		pthread_cond_signal( &condition );
		pthread_mutex_unlock( &condition_mutex );
	}

	static void *captureThread( void *p ) {
		Frame *frame = NULL;

		g_nav_ctl.capture_active  = TRUE;

		// Make sure we can tell the difference between a used and unused frame
		for (int index = 0; index < BUFFERED_FRAMES; index ++ )
			frameBuffer[ index ] = NULL;

		// Loop until we're informed otherwise
		while ( g_nav_ctl.capture_active  ) 
		{
			// Wait for the reader to indicate that something has happened
			reader->WaitForAction( );

			// Get the next frame
           	frame = reader->GetFrame();

			// Check if the out queue is falling behind
			bool critical_mass = reader->GetOutQueueSize( ) > reader->GetInQueueSize( );

			// Make sure we return the oldest frame first
			if ( frame != NULL && frameCount == BUFFERED_FRAMES ) 
			{
				int oldest = ( framePosition + 1 ) % BUFFERED_FRAMES;
				frameBuffer[ oldest ] = NULL;
				frameCount --;
			}

			// Put the last collected frame into the buffer so it can be picked up
			// by the video thread.
			if ( frame != NULL && !critical_mass ) 
			{
				// Bung it in to the buffer
				int newest = ( framePosition + 1 ) % BUFFERED_FRAMES;
				frameBuffer[ newest ] = frame;
				frameCount ++;
				framePosition = newest;
				TriggerAction( );
			}

			// Do with it what needs to be done
           	if (frame != NULL)
			{
				// Play audio
				if ( audioOn && !critical_mass )
					if (writer == NULL || (writer != NULL && Preferences::getInstance().preview_capture) )
	              		displayer->PutSound(*frame);

				// All access to the writer is protected
				pthread_mutex_lock( &writerlock );
               	if (writer != NULL && frame->IsNormalSpeed() )
				{
                   	if ( !writer->WriteFrame(*frame) )
					{
						pthread_mutex_unlock( &writerlock );
						gdk_threads_enter();
						common->getPageCapture()->stopCapture();
						common->setStatusBar( "Capture stopped. Disk Full?" );
						gdk_threads_leave();
					}
				}
				pthread_mutex_unlock( &writerlock );

				// Return the frame for reuse
				reader->DoneWithFrame( frame );
			}
		}

		// Set the buffer positional vars to defaults to allow restart
		frameCount = 0;
		framePosition = -1;

		g_nav_ctl.capture_active  = FALSE;

		TriggerAction( );

		return NULL;
	}


	static void *videoThread( gpointer p ) {
		int lastPosition = -1;

		g_nav_ctl.capture_active  = TRUE;
        GtkWidget *drawingarea = (GtkWidget*)p;

		while ( g_nav_ctl.capture_active  ) {

			int position = WaitForAction( lastPosition );

			// Lock the gui
			gdk_threads_enter();

			// If we have a valid position and it's different to the last iteration
			if ( g_nav_ctl.capture_active && position != -1 && position != lastPosition ) {

				// ... display it - note, the displayer and the frame may be destroyed
				// while waiting for the thread lock so additional checks are needed
				if ( ( displayer != NULL && frameBuffer[ position ] != NULL ) ) {
			    	if ( writer == NULL || (writer != NULL && Preferences::getInstance().preview_capture) ) {
	       				displayer->Put(*frameBuffer[ position ], drawingarea, TRUE);
					}
	       		}

				// Remember the buffer position of the last frame displayed
				lastPosition = position;
			}

			// Apply the changes to the GUI
			common->getPageCapture()->applyAVCState( avcStatus );
            common->showFrameInfo(0);
			gdk_flush();
			gdk_threads_leave();
		}

		return NULL;
	}

	static void *avcThread( gpointer p ) {
		g_nav_ctl.capture_active  = TRUE;

		while ( g_nav_ctl.capture_active  ) {
			// Check that device is still valid
			common->getPageCapture()->CheckDevices( );

			// Release temporarily
			struct timespec t = { 0,  (unsigned long) Preferences::getInstance().avcPollIntervalMs 
				* 1000000UL };
			nanosleep( &t, NULL );
		}

		return NULL;
	}

    gboolean
    on_capture_drawingarea_expose_event (GtkWidget *widget,
                                         GdkEventExpose *event,
                                         gpointer user_data) {

		windowChanged();
        return FALSE;
    }


    void
    on_capture_page_record_button_clicked (GtkButton *button,
                                            gpointer user_data) {
		startCapture();
    }


    void
    on_capture_page_stop_button_clicked (GtkButton *button,
                                           gpointer user_data) {
		stopCapture();
    }

	void
	on_capture_page_snapshot_button_clicked (GtkButton       *button,
	                                        gpointer         user_data)
	{
		saveFrame();
	}

	gboolean
	on_capture_page_file_entry_focus_in_event (GtkWidget       *widget,
	                                        GdkEventFocus   *event,
	                                        gpointer         user_data)
	{
		gtk_entry_set_editable( GTK_ENTRY( widget ), TRUE );
		g_nav_ctl.escaped = TRUE;
		strncpy( entry_text, gtk_entry_get_text( GTK_ENTRY( widget )), 1024 );
	  	return FALSE;
	}
	
	
	gboolean
	on_capture_page_file_entry_focus_out_event (GtkWidget       *widget,
	                                        GdkEventFocus   *event,
	                                        gpointer         user_data)
	{
		if (strcmp( gtk_entry_get_text( GTK_ENTRY( widget ) ), entry_text ) != 0)
		{
			GnomeFileEntry *fentry = GNOME_FILE_ENTRY( lookup_widget(common->getWidget(), "capture_page_gnome_file_entry") );
			GnomeEntry *entry = GNOME_ENTRY( gnome_file_entry_gnome_entry( fentry ) );
			gnome_entry_append_history( entry, FALSE, entry_text );
		}
		gtk_entry_set_editable( GTK_ENTRY( widget ), FALSE );
		g_nav_ctl.escaped = FALSE;
	  	return FALSE;
	}
	
	
	void
	on_capture_page_file_entry_grab_focus  (GtkWidget       *widget,
	                                        gpointer         user_data)
	{
		gtk_entry_set_editable( GTK_ENTRY( widget ), TRUE );
		g_nav_ctl.escaped = TRUE;
	}
	
	gboolean
	on_capture_page_file_entry_button_press_event (GtkWidget       *widget,
	                                        GdkEventButton  *event,
	                                        gpointer         user_data)
	{
		g_nav_ctl.escaped = TRUE;
		gtk_entry_set_editable( GTK_ENTRY(lookup_widget( widget, "capture_page_file_entry" )), TRUE );
		return FALSE;
	}

	void
	on_capture_page_gnome_file_entry_browse_clicked
	                                        (GnomeFileEntry  *gnomefileentry,
	                                        gpointer         user_data)
	{
		gtk_entry_set_editable( GTK_ENTRY( gnome_file_entry_gtk_entry( gnomefileentry ) ), TRUE );
		g_nav_ctl.escaped = TRUE;
	}



	static bool buttonGuard = true;
	void
	on_capture_page_button_pressed         (GtkButton       *button,
                                        	gpointer         user_data)
	{
		buttonGuard = false;
	}


	void
	on_capture_page_button_released        (GtkButton       *button,
                                        	gpointer         user_data)
	{
		buttonGuard = true;
	}


	void
	on_capture_page_mute_button_toggled    (GtkToggleButton *togglebutton,
                                        	gpointer         user_data)
	{
		if ( ! buttonGuard ) {
			audioOn = ! audioOn;
			buttonGuard = true;
			gtk_toggle_button_set_active( togglebutton, !audioOn );
		}
	}

	void
    on_capture_page_options_button_clicked (GtkButton *button,
            								gpointer user_data) {
		avcStatus = !avcStatus;
		common->getPageCapture( )->avc_enabled = !common->getPageCapture( )->avc_enabled;
		cout << ">> AV/C " << ( common->getPageCapture( )->avc_enabled ? "Enabled" : "Disabled" ) << endl;
		TriggerAction( );
    }

}
	
