/*
 * page_timeline.cc Notebook Timeline Page Object
 * Copyright (C) 2001 Charles Yates <charles.yates@pandora.be>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <pthread.h>

#include "page_timeline.h"
#include "page_editor.h"

/*
 * Callbacks
 */

extern "C" {

	extern KinoCommon *common;

	int getOneSecond(void)
	{
		Frame &frame = *( GetFramePool()->GetFrame( ) );
		common->getPlayList()->GetFrame(common->g_currentFrame, frame);
		int value = (frame.IsPAL() ? 25 : 30);
		GetFramePool( )->DoneWithFrame( &frame );
		return value;
	}

	static unsigned char	  pixels[FRAME_MAX_WIDTH*FRAME_MAX_HEIGHT*4];

	void selectIcon( GtkWidget *widget, int icon )
	{
		gpointer ptr = gnome_icon_list_get_icon_data( GNOME_ICON_LIST( widget ), icon );
		int val = 0;
		// Get the frame and move to it
		sscanf( (char *)ptr, "%d", &val );
		common->moveToFrame( val-1 );
		// Go to editor window
		common->changePageRequest( 0 );
	}

	void
	on_iconlist1_select_icon               (GnomeIconList   *gnomeiconlist,
                                        	gint             arg1,
                                        	GdkEvent        *event,
                                        	gpointer         user_data)
	{
		selectIcon( GTK_WIDGET( gnomeiconlist ), arg1 );
	}

	void
	on_timeline_ok_button_pressed          (GtkButton       *button,
                                        	gpointer         user_data)
	{
		common->getPageTimeline( )->showIcons(  );
	}

	static gboolean on_iconlist_refresh_required( GtkWidget *some_widget, void *some_event, gpointer user_data )
	{
		( ( PageTimeline *)user_data )->refresh( );
		return false;
	}
}

PageTimeline::PageTimeline( KinoCommon *common ) : refresh_required(false), action(0),
	last_begin(0), last_end(0), scene(-1) 
{ 
	this->common = common;
	pthread_mutex_init( &key, NULL );

	GtkWidget *widget = lookup_widget( common->getWidget( ), "iconlist1" );
	gtk_signal_connect( GTK_OBJECT( widget ), "event", GTK_SIGNAL_FUNC( on_iconlist_refresh_required ), this );
}

void PageTimeline::start() {
	cout << ">> Starting timeline" << endl;

	if ( common->getPlayList()->GetNumFrames() > 0 )
	{
		int begin = common->getPlayList()->FindStartOfScene( common->g_currentFrame );
		int end = common->getPlayList()->FindEndOfScene( common->g_currentFrame );

		if ( last_begin != begin || last_end != end )
		{
			GtkEntry *startSpin = GTK_ENTRY( lookup_widget( common->getWidget(), "start_spin" ) );
			GtkEntry *endSpin = GTK_ENTRY( lookup_widget( common->getWidget(), "end_spin" ) );	
			GtkAdjustment *adjust = gtk_spin_button_get_adjustment( GTK_SPIN_BUTTON( startSpin ) );
	
			adjust->lower = 1;
			adjust->upper = common->getPlayList()->GetNumFrames() + 1;
			gtk_spin_button_set_value( GTK_SPIN_BUTTON( startSpin ), begin + 1);
			gtk_signal_emit_by_name( GTK_OBJECT( adjust ), "changed" );

			adjust = gtk_spin_button_get_adjustment( GTK_SPIN_BUTTON( endSpin ) );
			adjust->lower = 0;
			adjust->upper = common->getPlayList()->GetNumFrames() + 1;
			gtk_spin_button_set_value( GTK_SPIN_BUTTON( endSpin ), end + 1 );
			gtk_signal_emit_by_name( GTK_OBJECT( adjust ), "changed" );

			last_begin = begin;
			last_end = end;

			scene = 0;
			vector <int> scenes = common->getPageEditor()->GetScene();
			while ( begin > scenes[ scene++ ] );
			scene--;
			
			refresh_required = true;
		}
	}
}

void PageTimeline::selectScene( int i ) {
	cout << "Time line scene " << i << endl;
	GtkWidget *start_spin = lookup_widget( common->getWidget(), "start_spin" );
	GtkWidget *end_spin = lookup_widget( common->getWidget(), "end_spin" );
	GtkAdjustment *adjust = gtk_spin_button_get_adjustment( GTK_SPIN_BUTTON( start_spin ) );
	vector <int> scenes = common->getPageEditor()->GetScene();

	if ( i < 0 )
		i = 0;
	if ( i >= (int) scenes.size() )
		i = scenes.size() - 1;
	int value = i == 0 ? 0 : scenes[ i - 1 ];
	adjust->lower = 1;
	adjust->upper = scenes[ scenes.size() - 1 ] + 1;
	gtk_spin_button_set_value( GTK_SPIN_BUTTON( start_spin ), value + 1);
	gtk_signal_emit_by_name( GTK_OBJECT( adjust ), "changed" );

	adjust = gtk_spin_button_get_adjustment( GTK_SPIN_BUTTON( end_spin ) );

	adjust->lower = 1;
	adjust->upper = scenes[ scenes.size() - 1 ] + 1;
	gtk_spin_button_set_value( GTK_SPIN_BUTTON( end_spin ), scenes[ i ] );
	gtk_signal_emit_by_name( GTK_OBJECT( adjust ), "changed" );

	last_begin = value;
	last_end = scenes[ i ] - 1;
	scene = i;

	common->g_currentFrame = value;

	Frame &frame = *( GetFramePool( )->GetFrame( ) );
	FileHandler *media;
	common->getPlayList()->GetMediaObject( value, &media );
	common->getPlayList()->GetFrame( value, frame );
	common->showFrameMoreInfo( frame, media );
	GetFramePool( )->DoneWithFrame( &frame );
	common->setCurrentScene( value );
	showIcons( );
}

gulong PageTimeline::activate() {
	return SCENE_LIST;
}

void PageTimeline::showIcons( ) 
{
	pthread_mutex_lock( &key );
	int my_action = ++ action;

	GtkWidget *list = lookup_widget( GTK_WIDGET( common->getWidget() ), "iconlist1" );

	gnome_icon_list_set_row_spacing( GNOME_ICON_LIST( list ), 4 );
	gnome_icon_list_set_text_spacing( GNOME_ICON_LIST( list ), 0 );

	Frame &frame = *GetFramePool( )->GetFrame( );
	GdkImlibImage *image;
	GdkImlibImage *scaled;
	GtkWidget *start_spin = lookup_widget( list, "start_spin" );
	GtkWidget *end_spin = lookup_widget( list, "end_spin" );
	char label[ 256 ] = "";
	frame.decoder->quality = DV_QUALITY_DC;
	if (Preferences::getInstance().displayQuality < 4)
		frame.decoder->quality |= DV_QUALITY_COLOR;

	int start = atoi( gtk_entry_get_text( GTK_ENTRY( start_spin ) ) );
	int end = atoi( gtk_entry_get_text( GTK_ENTRY( end_spin ) ) );
	int items_per_row = gnome_icon_list_get_items_per_line( GNOME_ICON_LIST( list ) );
	int number_of_rows = list->allocation.height / 76;
	int total_possible = items_per_row * number_of_rows;
	int total_required = end - start + 1;
	double increment = 1;

	if ( total_required == 1 )
		increment = 2;
	else if ( total_possible < total_required )
		increment = 1 / ( (double) total_possible - 1 );
	else
		increment = 1 / ( (double) total_required - 1 );

	gnome_icon_list_freeze( GNOME_ICON_LIST( list ) );
	gnome_icon_list_clear( GNOME_ICON_LIST( list ) );
	gnome_icon_list_thaw( GNOME_ICON_LIST( list ) );

	vector< GdkImlibImage * >::iterator iter;
	for ( iter = icons.begin(); iter != icons.end(); iter++ )
		gdk_imlib_destroy_image( *iter );

	icons.erase( icons.begin(), icons.end() );
	start--;
	end--;
	for ( double position = 0; action == my_action && position <= 1; position += increment )
	{
		int index = start + (int)( position * total_required );
		gnome_icon_list_freeze( GNOME_ICON_LIST( list ) );
		common->getPlayList()->GetFrame((int)index, frame);
		frame.ExtractRGB(pixels);
		image = gdk_imlib_create_image_from_data( pixels, NULL, frame.GetWidth(), frame.GetHeight() );
		scaled = gdk_imlib_clone_scaled_image( image, 90, 72 );
		icons.push_back( scaled );
		sprintf( label, "%d", (int)index+1 );
		int pos = gnome_icon_list_append_imlib( GNOME_ICON_LIST(list), scaled, "" );
		gnome_icon_list_set_icon_data( GNOME_ICON_LIST(list), pos, strdup( label ) );
		gdk_imlib_destroy_image( image );
		gnome_icon_list_thaw( GNOME_ICON_LIST( list ) );
	    while ( action == my_action && gtk_events_pending()) 
		{
			pthread_mutex_unlock( &key );
       		gtk_main_iteration();
			pthread_mutex_lock( &key );
		}
	}
	GetFramePool( )->DoneWithFrame( &frame );
	pthread_mutex_unlock( &key );
}

void PageTimeline::refresh( )
{
	if ( refresh_required )
	{
		refresh_required = false;
		showIcons( );
	}
}

gboolean PageTimeline::processKeyboard( GdkEventKey *event ) {
	gboolean ret = FALSE;

		// Translate special keys to equivalent command
		switch (event->keyval) {
			case GDK_k:
			case GDK_Up:
				selectScene( scene - 1 );
				break;
			case GDK_j:
			case GDK_Return:
			case GDK_Down:
				selectScene( scene + 1 );
				break;
			case GDK_Escape:
				common->changePageRequest( PAGE_EDITOR );
				break;
			default:
				break;
		}
	
	return ret;
}
