/*
 * page_edit.h Notebook Trim Page Object
 * Copyright (C) 2001 Dan Dennedy <dan@dennedy.org>
 * Custom GTK+ Trim Control by Ronald Bultje
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
using std::cout;
using std::endl;

#include <sys/time.h>
#include <pthread.h>

#include "page_trim.h"
#include "avi.h"
#include "error.h"
#include "message.h"
#include "filehandler.h"
#include "frame.h"
#include "commands.h"
#include "page_editor.h"
#include "ieee1394io.h"

#undef PLAY_WITH_STATS

#define TRIM_ADJ_POS 0
#define TRIM_ADJ_IN 1
#define TRIM_ADJ_OUT 2

extern KinoCommon *common;

extern "C" {
	#include "support.h"
	extern struct navigate_control g_nav_ctl;
	extern char cmd[];
	static char lastcmd[ 256 ] = { 0 };
	static int _getOneSecond(void);

	static void resetThreads( );
	static void *readThread( void *info );
	static void *videoThread( void *info );
	static void *audioThread( void *info );
	static pthread_t readthreadTrim;
	static pthread_t audiothreadTrim;
	static pthread_t videothreadTrim;
	static pthread_mutex_t threadlock = PTHREAD_MUTEX_INITIALIZER;
	static GtkAdjustment *trim_adj[3];
	static gboolean skipPosUpdate = FALSE;
	static GtkSpinButton *spin_in;
	static GtkSpinButton *spin_out;
	static GtkToggleButton *link_toggle;
	static GtkToggleButton *loop_toggle;
	static gboolean doScrub = false;

	#define PLAYBACK_FRAMES 50
	static Frame *frameContent[ PLAYBACK_FRAMES ];
	extern dv1394Writer *dv1394;

	/* signal callbacks */
	gboolean
	on_trim_value_changed_event      (GtkWidget       *widget,
											GdkEventButton  *event,
											gpointer         user_data)
	{
		int value = (int) GTK_ADJUSTMENT(widget)->value;
		if (GTK_ADJUSTMENT(widget) == trim_adj[TRIM_ADJ_IN]) {
			int diff = common->getPageTrim()->getOutPoint() - common->getPageTrim()->getInPoint() + value;
			
			if (gtk_toggle_button_get_active(link_toggle) == TRUE && skipPosUpdate == FALSE) {
				if (diff <= common->getPageTrim()->getTotalFrames()) {
					if (value >= common->getPageTrim()->getOutPoint())
						gtk_adjustment_set_value( trim_adj[TRIM_ADJ_OUT], value+1);
					else
						common->getPageTrim()->setInPoint(value);
	
					skipPosUpdate = TRUE;
					gtk_spin_button_set_value( spin_out, (gfloat) diff + 1 );
					skipPosUpdate = TRUE;
					gtk_adjustment_set_value( trim_adj[TRIM_ADJ_OUT], diff);
				} else {
					skipPosUpdate = TRUE;
					gtk_adjustment_set_value( trim_adj[TRIM_ADJ_IN], common->getPageTrim()->getInPoint());
				}
			} else {
				if (value > common->getPageTrim()->getOutPoint())
					gtk_adjustment_set_value( trim_adj[TRIM_ADJ_OUT], value);
				else
					common->getPageTrim()->setInPoint(value);
			}
			skipPosUpdate = TRUE;
			if (common->g_currentFrame > -1) common->getPageTrim()->movedToFrame(value);
			gtk_spin_button_set_value(spin_in, (gfloat) value+1);

		} else if (GTK_ADJUSTMENT(widget) == trim_adj[TRIM_ADJ_OUT]) {
			int diff = common->getPageTrim()->getInPoint() - common->getPageTrim()->getOutPoint() + value;
			
			if (gtk_toggle_button_get_active(link_toggle) == TRUE && skipPosUpdate == FALSE) {
				if (diff >= 0) {
					if (value <= common->getPageTrim()->getInPoint())
						gtk_adjustment_set_value( trim_adj[TRIM_ADJ_IN], value-1);
					else
						common->getPageTrim()->setOutPoint(value);
		
					skipPosUpdate = TRUE;
					gtk_spin_button_set_value( spin_in, (gfloat) diff + 1);
					skipPosUpdate = TRUE;
					gtk_adjustment_set_value( trim_adj[TRIM_ADJ_IN], diff);
				} else {
					skipPosUpdate = TRUE;
					gtk_adjustment_set_value( trim_adj[TRIM_ADJ_OUT], common->getPageTrim()->getOutPoint());
				}
			} else {
				if (value < common->getPageTrim()->getInPoint())
					gtk_adjustment_set_value( trim_adj[TRIM_ADJ_IN], value);
				else
					common->getPageTrim()->setOutPoint(value);
			}
			skipPosUpdate = TRUE;
			if (common->g_currentFrame > -1) common->getPageTrim()->movedToFrame(value);
			gtk_spin_button_set_value(spin_out, (gfloat) value+1);

		} else if (skipPosUpdate == FALSE) {
			skipPosUpdate = TRUE;
			if (common->g_currentFrame > -1) common->getPageTrim()->movedToFrame(value);
		}
		skipPosUpdate = FALSE;
		return FALSE;
	}


	void
	on_spinbutton_trim_in_changed          (GtkEditable     *editable,
											gpointer         user_data)
	{
		int value;
		sscanf( gtk_entry_get_text( GTK_ENTRY( spin_in ) ), "%d", &value );
		gtk_adjustment_set_value( trim_adj[TRIM_ADJ_IN], value-1);	
		}
	
	
	void
	on_spinbutton_trim_out_changed         (GtkEditable     *editable,
											gpointer         user_data)
	{
		int value;
		sscanf( gtk_entry_get_text( GTK_ENTRY( spin_out ) ), "%d", &value );
		gtk_adjustment_set_value( trim_adj[TRIM_ADJ_OUT], value-1);
	}
	
	
	void
	on_button_trim_in_reset_clicked        (GtkButton       *button,
											gpointer         user_data)
	{
		common->getPageTrim()->resetInPoint();
	}
	
	
	void
	on_button_trim_out_reset_clicked       (GtkButton       *button,
											gpointer         user_data)
	{
		common->getPageTrim()->resetOutPoint();
	}


	void
	on_button_trim_in_set_clicked          (GtkButton       *button,
											gpointer         user_data)
	{
		int value = common->getPageTrim()->getPosition();
		gtk_adjustment_set_value( trim_adj[TRIM_ADJ_IN], value);	
	}
	
	
	void
	on_button_trim_out_set_clicked         (GtkButton       *button,
											gpointer         user_data)
	{
		int value = common->getPageTrim()->getPosition();
		gtk_adjustment_set_value( trim_adj[TRIM_ADJ_OUT], value);	
	}

	void
	on_togglebutton_trim_link_toggled      (GtkToggleButton *togglebutton,
											gpointer         user_data)
	{
	}

	gboolean
	on_spinbutton_trim_in_focus_in_event   (GtkWidget       *widget,
											GdkEventFocus   *event,
											gpointer         user_data)
	{
		g_nav_ctl.escaped = TRUE;
		return FALSE;
	}
	
	
	gboolean
	on_spinbutton_trim_in_focus_out_event  (GtkWidget       *widget,
											GdkEventFocus   *event,
											gpointer         user_data)
	{
		g_nav_ctl.escaped = FALSE;
		return FALSE;
	}
	
	
	gboolean
	on_spinbutton_trim_out_focus_in_event  (GtkWidget       *widget,
											GdkEventFocus   *event,
											gpointer         user_data)
	{
		g_nav_ctl.escaped = TRUE;
		return FALSE;
	}
	
	
	gboolean
	on_spinbutton_trim_out_focus_out_event (GtkWidget       *widget,
											GdkEventFocus   *event,
											gpointer         user_data)
	{
		g_nav_ctl.escaped = FALSE;
		return FALSE;
	}

	gboolean
	on_trim_button_press_event        (GtkWidget       *widget,
											GdkEventButton  *event,
											gpointer         user_data)
	{
		doScrub = TRUE;
		videoPause();
		return FALSE;
	}
	
	
	gboolean
	on_trim_button_release_event      (GtkWidget       *widget,
											GdkEventButton  *event,
											gpointer         user_data)
	{
		doScrub = FALSE;
		return FALSE;
	}
}

/** Constructor for the trimmer page object.
	
  	\param common	KinoCommon object to which this page belongs
*/
GArray *scenes = NULL;
PageTrim::PageTrim( KinoCommon *common ) : 
	start_orig(0), end_orig(0), in_orig(0), out_orig(0), in(0), out(0),
	max(0),	pos(-1), currentScene(0), lastPos(-1)
{
	cout << "> Creating page trim" << endl;
	this->common = common;
	this->frameArea = GTK_DRAWING_AREA( lookup_widget( common->getWidget(), "drawingarea_trim" ) );
	this->positionLabel = GTK_LABEL( lookup_widget( common->getWidget(), "position_label" ) );
	spin_in = GTK_SPIN_BUTTON( lookup_widget( common->getWidget(), "spinbutton_trim_in") );
	spin_out = GTK_SPIN_BUTTON( lookup_widget( common->getWidget(), "spinbutton_trim_out") );
	link_toggle = GTK_TOGGLE_BUTTON(lookup_widget( common->getWidget(), "togglebutton_trim_link") );
	loop_toggle = GTK_TOGGLE_BUTTON(lookup_widget( common->getWidget(), "togglebutton_trim_loop") );
	
	this->g_copiedPlayList = new PlayList();
	dv1394 = NULL;

	/* Attach the custom GTK+ Trim control from Ron Bultje */
	trim_adj[TRIM_ADJ_POS] = GTK_ADJUSTMENT(
		gtk_adjustment_new( 0, 0, 0, 1, 10, 0));
	trim_adj[TRIM_ADJ_IN] = GTK_ADJUSTMENT(
		gtk_adjustment_new( 0, 0, 0, 1, 10, 0));
	trim_adj[TRIM_ADJ_OUT] = GTK_ADJUSTMENT(
		gtk_adjustment_new( 0, 0, 0, 1, 10, 0));
	for (int i = 0; i < 3; i++)
		gtk_signal_connect (GTK_OBJECT (trim_adj[i]), "value_changed",
			GTK_SIGNAL_FUNC (on_trim_value_changed_event), (gpointer) i);
	trim = gtk_enhanced_scale_new( (GtkObject**) trim_adj, 3);
	gtk_widget_set_name (trim, "trim");
	gtk_widget_ref(trim);
	gtk_object_set_data_full (GTK_OBJECT(common->getWidget()), "trim", trim,
						(GtkDestroyNotify) gtk_widget_unref);
	GtkWidget *vbox_trim = lookup_widget(common->getWidget(), "vbox_trim");
	gtk_widget_show(trim);
	gtk_box_pack_start(GTK_BOX (vbox_trim), trim, FALSE, TRUE, 0);
	gtk_signal_connect( GTK_OBJECT(trim), "button_press_event",
		GTK_SIGNAL_FUNC(on_trim_button_press_event), NULL );
	gtk_signal_connect( GTK_OBJECT(trim), "button_release_event",
		GTK_SIGNAL_FUNC(on_trim_button_release_event), NULL );
}

/** Destructor for the trimmer page object.
*/

PageTrim::~PageTrim( ) {
	cout << "> Destroying page trim" << endl;
	delete this->g_copiedPlayList;
}

/** New File action.
*/

void PageTrim::newFile() {
	this->stopNavigator();
}

/** Start action. Called when the page becomes current.
*/

void PageTrim::start() {
	cout << ">> Starting Trimmer" << endl;

	start_orig = end_orig = in = in_orig = out = out_orig = max = 0;
	
	for ( int i = 0; i < PLAYBACK_FRAMES; i ++ )
		frameContent[ i ] = GetFramePool( )->GetFrame( );

	this->displayer = new FrameDisplayer();
	if ( Preferences::getInstance().dvExportDevice > 0 
			&& Preferences::getInstance().dv1394Preview )
		dv1394 = new dv1394Writer(
			Preferences::getInstance().dvExportDevice,
			Preferences::getInstance().channel,
			Preferences::getInstance().dvExportBuffers,
			Preferences::getInstance().cip_n,
			Preferences::getInstance().cip_d,
			Preferences::getInstance().syt_offset
		);
	
	if (common->g_currentFrame > -1) {
		if (common->getPlayList()->GetNumFrames() > 0) {
			currentScene = common->g_currentFrame;
			loadScene( currentScene );
			lastPos = pos = in + common->g_currentFrame - start_orig;
			
			trim_adj[TRIM_ADJ_IN]->upper = max;
			trim_adj[TRIM_ADJ_IN]->value = in;
			gtk_signal_emit_by_name( GTK_OBJECT(trim_adj[TRIM_ADJ_IN]), "changed" );
			trim_adj[TRIM_ADJ_OUT]->upper = max;
			trim_adj[TRIM_ADJ_OUT]->value = out;
			gtk_signal_emit_by_name( GTK_OBJECT(trim_adj[TRIM_ADJ_OUT]), "changed" );
			trim_adj[TRIM_ADJ_POS]->upper = max;
			trim_adj[TRIM_ADJ_POS]->value = pos;
			gtk_signal_emit_by_name( GTK_OBJECT(trim_adj[TRIM_ADJ_POS]), "changed" );
			
			GtkAdjustment *adjust = gtk_spin_button_get_adjustment( spin_in);
			adjust->lower = 1;
			adjust->upper = max  + 1;
			gtk_signal_emit_by_name( GTK_OBJECT( adjust ), "changed" );
			gtk_spin_button_set_value(spin_in, (gfloat) in + 1);

			adjust = gtk_spin_button_get_adjustment( spin_out);
			adjust->lower = 1;
			adjust->upper = max + 1;
			gtk_signal_emit_by_name( GTK_OBJECT( adjust ), "changed" );
			gtk_spin_button_set_value(spin_out, (gfloat) out + 1);
		}
	}
	else
	{
		common->loadSplash(this->frameArea);
	}
	
	common->packIt( "packer_trim" );
}

/** Activate the returned widgets. 
*/

gulong PageTrim::activate() {
	return COMMAND_MENU | 
		   SCENE_LIST |
		   VIDEO_START_OF_MOVIE |
		   VIDEO_START_OF_SCENE |
		   VIDEO_REWIND |
		   VIDEO_BACK |
		   VIDEO_PLAY |
		   VIDEO_PAUSE |
		   VIDEO_STOP |
		   VIDEO_FORWARD |
		   VIDEO_FAST_FORWARD |
		   VIDEO_NEXT_SCENE |
		   VIDEO_END_OF_MOVIE |
		   VIDEO_COMMAND |
		   VIDEO_SHUTTLE |
		   INFO_FRAME;
}

/** Clean action. Called when another page becomes current.
*/

void PageTrim::clean() {
	cout << ">> Leaving Trimmer" << endl;
	stopNavigator();
	delete this->displayer;
	if (dv1394 != NULL)
		delete dv1394;
	saveScene();
	for ( int i = 0; i < PLAYBACK_FRAMES; i ++ )
		GetFramePool( )->DoneWithFrame( frameContent[ i ] );

	cout << ">> Left Trimmer" << endl;
}


void PageTrim::loadScene( int currentScene )
{
	char value[20];
	PlayList newPlayList;
	playlist = newPlayList;
	PlayList *pl = common->getPlayList();
	FileHandler *media;

	start_orig = pl->FindStartOfScene( currentScene );
	end_orig = pl->FindEndOfScene( currentScene );
	in = in_orig = pl->GetClipBegin( currentScene );

	pl->GetPlayList( start_orig, end_orig, playlist );
	playlist.GetMediaObject( playlist.FindEndOfScene(0), &media );
	playlist.SetClipBegin(0, "0");
	snprintf( value, 19, "%i", media->GetTotalFrames()-1 );
	playlist.SetClipEnd( 0, value );

	max = playlist.GetNumFrames() -1;
	out = out_orig = max - (media->GetTotalFrames()-1 - pl->GetClipEnd(end_orig));
	common->setCurrentScene( start_orig );

}


void PageTrim::saveScene()
{
	if ( playlist.GetNumFrames() && ( in_orig != in || out_orig != out ) )
	{
		playlist.Delete(out+1, max);
		playlist.Delete(0, in-1);
		common->getPlayList()->Delete(start_orig, end_orig);
		common->getPlayList()->InsertPlayList(playlist, start_orig);
		GetEditorBackup()->Store( common->getPlayList() );
		common->hasListChanged = TRUE;
		common->setWindowTitle( );
	}
}


/** Starts the navigator thread. See comments on g_nav_ctl.
*/

void PageTrim::startNavigator() {
	stopNavigator( );
	g_nav_ctl.active = TRUE;
	resetThreads( );
	pthread_create( &readthreadTrim, NULL, readThread, &g_nav_ctl );
	pthread_create( &audiothreadTrim, NULL, audioThread, &g_nav_ctl );
	pthread_create( &videothreadTrim, NULL, videoThread, (gpointer) &g_nav_ctl);
}

/** Stops the navigator thread. See comments on g_nav_ctl.
*/

void PageTrim::stopNavigator() {
	if ( g_nav_ctl.active ) {
		g_nav_ctl.active = FALSE;
		gdk_threads_leave();
		pthread_join( readthreadTrim, NULL );
		pthread_join( audiothreadTrim, NULL );
		pthread_join( videothreadTrim, NULL );
		gdk_threads_enter();
		common->moveToFrame();
	}
}

/** Called when the current frame has changed through the common moveToFrame
	method and this page is the current page.

	\param	frame	frame moved to
*/

void PageTrim::movedToFrame( int frame ) {
	if (max > 0) {
		if ( frame > out && gtk_toggle_button_get_active(loop_toggle) == TRUE )
			pos = in;
		else if ( frame < in && gtk_toggle_button_get_active(loop_toggle) == TRUE )
			pos = out;
		else if ( frame >= 0 && frame <= max )
			pos = frame;
		else if ( frame >= max )
			pos = max;
		else if ( frame < 0 && max > 0 )
			pos = 0;
		else
			pos = -1;
	
		if ( g_nav_ctl.active == FALSE) {
			skipPosUpdate = FALSE;
			showFrame( pos, (pos==lastPos) || (Preferences::getInstance().audioScrub==FALSE) );
		}
		else {
			stopNavigator( );
			skipPosUpdate = FALSE;
			showFrame( pos, (pos==lastPos) || (Preferences::getInstance().audioScrub==FALSE) );
			startNavigator( );
		}
	}
}

/** Show the frame requested

  	\param i		frame to be shown
	\param no_audio	indicate if audio is required or not
*/

void PageTrim::showFrame( int i, gboolean no_audio ) {
	if (max == 0) {
		common->loadSplash( frameArea );
		common->showFrameInfo( -1 );
		return;
	}
	else if ( i > -1 )
	{
		Frame &frame = *( GetFramePool( )->GetFrame( ) );
		if (i > max) i = max;
		playlist.GetFrame(i, frame);
		common->showFrameInfo( i );
		if (dv1394 != NULL)
			dv1394->SendFrame( frame );
		getFrameDisplayer()->Put(frame, GTK_WIDGET( frameArea ), no_audio);
		lastPos = i;
		skipPosUpdate = TRUE;
		gtk_adjustment_set_value(trim_adj[TRIM_ADJ_POS], (gfloat) i);
		GetFramePool( )->DoneWithFrame( &frame );
	}
}

void PageTrim::showFrame( int position, Frame& frame ) {
	if ( max == 0 ) {
		common->loadSplash( frameArea );
	}
	else {
		getFrameDisplayer()->Put( frame, GTK_WIDGET( frameArea ), TRUE );
		common->showFrameInfo( position );
		skipPosUpdate = TRUE;
		gtk_adjustment_set_value(trim_adj[TRIM_ADJ_POS], (gfloat) position);
	}
}


/** Move to the start of the scene.
*/

void PageTrim::videoStartOfMovie() {
	movedToFrame( 0 );
	common->toggleComponents( VIDEO_START_OF_MOVIE, false);
}

/** Move to the in point of the scene.
*/

void PageTrim::videoPreviousScene() {
	movedToFrame( in );
	common->toggleComponents( VIDEO_START_OF_SCENE, false);
}

/** Move to the in point of the scene.
*/

void PageTrim::videoStartOfScene() {
	movedToFrame( in );
	common->toggleComponents( VIDEO_START_OF_SCENE, false);
}

/** Rewind.
*/

void PageTrim::videoRewind() {
	common->toggleComponents( common->getComponentState(), false );

	// Toggle Rewind state
	if ( g_nav_ctl.step != -10 ) {
		common->toggleComponents( VIDEO_REWIND, true );
		g_nav_ctl.step = -10;
		startNavigator();
	}
	else {
		stopNavigator();
		g_nav_ctl.step = 1;
		common->toggleComponents( VIDEO_REWIND, false );
		common->toggleComponents( VIDEO_STOP, true );
	}
}

/** Move one frame back. If the navigator is active, then this action toggles 
	between reverse and stop.
*/

void PageTrim::videoBack() {
	common->toggleComponents( common->getComponentState(), false );
	if ( g_nav_ctl.active ) {
		if ( g_nav_ctl.step != -1 ) {
			g_nav_ctl.step = -1;
			common->toggleComponents( VIDEO_BACK, true );
			startNavigator();
		}
		else {
			stopNavigator();
			g_nav_ctl.step = 1;
			common->toggleComponents( VIDEO_BACK, false );
			common->toggleComponents( VIDEO_STOP, true );
		}
	} 
	else {
		movedToFrame( pos-1 );
		common->toggleComponents( VIDEO_BACK, false );
		common->toggleComponents( VIDEO_STOP, true );
	}
}

/** Play.
*/

void PageTrim::videoPlay() {
	common->toggleComponents( common->getComponentState(), false );
	if ( g_nav_ctl.active == FALSE || g_nav_ctl.step != 1 ) {
		common->toggleComponents( VIDEO_PLAY, true );
		g_nav_ctl.step = 1;
		startNavigator();
	}
	else {
		stopNavigator();
		common->toggleComponents( VIDEO_PLAY, false );
		common->toggleComponents( VIDEO_STOP, true );
	}
}

/** Pause.
*/

void PageTrim::videoPause() {
	common->toggleComponents( common->getComponentState(), false );
	common->toggleComponents( VIDEO_PAUSE, true );
	stopNavigator();
}

/** Stop.
*/

void PageTrim::videoStop() {
	common->toggleComponents( common->getComponentState(), false );
	common->toggleComponents( VIDEO_STOP, true );
	stopNavigator();
}

/** Move one frame forward. If the navigator is active, then this action puts
	the video into normal play.
*/

void PageTrim::videoForward() {
	common->toggleComponents( common->getComponentState(), false );
	if ( g_nav_ctl.active && g_nav_ctl.step != 1 ) {
		g_nav_ctl.step = 1;
		common->toggleComponents( VIDEO_FORWARD, true );
		startNavigator();
	} else {
		stopNavigator();
		movedToFrame( pos+1 );
		common->toggleComponents( VIDEO_FORWARD, false);
		common->toggleComponents( VIDEO_STOP, true);
	}
}

/** Fast forward.
*/

void PageTrim::videoFastForward() {
	common->toggleComponents( common->getComponentState(), false );

	if ( g_nav_ctl.step != 10 ) {
		common->toggleComponents( VIDEO_FAST_FORWARD, true );
		g_nav_ctl.step = 10;
		startNavigator();
	}
	else {
		stopNavigator();
		g_nav_ctl.step = 1;
		common->toggleComponents( VIDEO_FAST_FORWARD, false );
		common->toggleComponents( VIDEO_STOP, true );
	}
}

/** Shuttle

	Bi-directionaly variable-speed playback.
	
	\param angle A number from -7 (fastest reverse) to 7 (fastest forward), 0=stop
*/

void PageTrim::videoShuttle( int angle ) {
	int speedTable[] = { 0, 20, 50, 100, 200, 500, 1000, (_getOneSecond() * 100) };
	char s[64];

	stopNavigator( );

	if (angle < -7) angle = -7;
	if (angle > 7) angle = 7;

	int speed = speedTable[(angle < 0) ? -angle : angle] * ((angle < 0) ? -1 : 1);
	if (speed == 0) {
		common->keyboardFeedback( "", "" );
		common->toggleComponents( VIDEO_PAUSE, true );
		return;
	}

	g_nav_ctl.step = speed / 100;
	g_nav_ctl.rate = 100 / speed;
	g_nav_ctl.subframe = 0;
	
	snprintf( s, 63, "Shuttle %+.1f fps", (float) speed / 100.0 * (float) _getOneSecond() );
	common->keyboardFeedback( "", s );
	
	startNavigator();
}

/** Move to the end of the current scene.
*/

void PageTrim::videoNextScene() {
	movedToFrame( out );
	common->toggleComponents( VIDEO_NEXT_SCENE, false);
}

/** Move to the end of the current scene.
*/

void PageTrim::videoEndOfScene() {
	movedToFrame( out );
	common->toggleComponents( VIDEO_NEXT_SCENE, false);
}

/** Move to the end of the play list.
*/

void PageTrim::videoEndOfMovie() {
	common->toggleComponents( common->getComponentState(), false );
	common->toggleComponents( VIDEO_END_OF_MOVIE, true);
	stopNavigator();
	movedToFrame( max );
	common->toggleComponents( VIDEO_END_OF_MOVIE, false);
	common->toggleComponents( VIDEO_STOP, true);
}

/** Load the selected scene.

	\param i	scene to move to
*/

void PageTrim::selectScene( int i ) {

	saveScene();
	common->getPageEditor()->DrawBar( common->g_currentFrame );

	vector <int> scene = common->getPageEditor()->GetScene();
	if (i >= (int) scene.size()) i = scene.size() -1;
	currentScene = (i <= 0) ? 0 : scene[ i - 1 ];

	common->g_currentFrame = currentScene;

	/* figure out what scene I am in */
	loadScene( currentScene );
	lastPos = pos = in;

	trim_adj[TRIM_ADJ_IN]->upper = max;
	gtk_signal_emit_by_name( GTK_OBJECT(trim_adj[TRIM_ADJ_IN]), "changed" );
	trim_adj[TRIM_ADJ_OUT]->upper = max;
	gtk_signal_emit_by_name( GTK_OBJECT(trim_adj[TRIM_ADJ_OUT]), "changed" );
	trim_adj[TRIM_ADJ_POS]->upper = max;
	gtk_signal_emit_by_name( GTK_OBJECT(trim_adj[TRIM_ADJ_POS]), "changed" );

	GtkAdjustment *adjust = gtk_spin_button_get_adjustment( spin_in);
	adjust->lower = 1;
	adjust->upper = max + 1;
	gtk_signal_emit_by_name( GTK_OBJECT( adjust ), "changed" );

	adjust = gtk_spin_button_get_adjustment( spin_out);
	adjust->lower = 1;
	adjust->upper = max + 1;
	gtk_signal_emit_by_name( GTK_OBJECT( adjust ), "changed" );

	gtk_adjustment_set_value( trim_adj[TRIM_ADJ_IN], (gfloat) in);
	gtk_adjustment_set_value( trim_adj[TRIM_ADJ_OUT], (gfloat) out);

	movedToFrame(in);

}

int PageTrim::getSceneIndex() {
	unsigned int i = 0;
	vector <int> scene = common->getPageEditor()->GetScene();
	for (i = 0; i < scene.size()-1; i++)
		if (currentScene < scene[i]) break;
	return i;
}

void PageTrim::resetInPoint() {
	in = in_orig;
	gtk_adjustment_set_value( trim_adj[TRIM_ADJ_IN], (gfloat) in);
	gtk_spin_button_set_value(spin_in, (gfloat) in + 1);
}


void PageTrim::resetOutPoint() {
	out = out_orig;
	gtk_adjustment_set_value( trim_adj[TRIM_ADJ_OUT], (gfloat) out);
	gtk_spin_button_set_value(spin_out, (gfloat) out + 1);
}

/** Process a keyboard event. 

  	\param event	keyboard event
*/

gboolean PageTrim::processKeyboard( GdkEventKey *event ) {
	gboolean ret = FALSE;

	// Only process while not escape mode
	if ( g_nav_ctl.escaped == FALSE ) {

		// Translate special keys to equivalent command
		switch (event->keyval) {
			case GDK_BackSpace:
			case GDK_Left:
				strcat(cmd, "h");
				break;
			case GDK_Up:
				strcat(cmd, "k");
				break;
			case GDK_Right:
				strcat(cmd, "l");
				break;
			case GDK_Return:
			case GDK_Down:
				strcat(cmd, "j");
				break;
			case GDK_Delete:
				strcat( cmd, "x" );
				break;
			case GDK_Escape:
				if (common->getComponentState() & VIDEO_STOP) {
					common->changePageRequest( PAGE_EDITOR );
				} else {
					common->keyboardFeedback(cmd, "Stop");
					common->videoStop( );
					cmd[0] = 0;
				}
				return FALSE;
			default:
				if ( strcmp( event->string, "." ) )
					strcat(cmd, event->string);
				break;
		}
	
		bool commandLine = false;
		if ( !strcmp( event->string, "." ) ) {
			strcpy( cmd, lastcmd );
			commandLine = lastcmd[ 0 ] == ':' && strlen( lastcmd ) > 1;
		}

#if 0
		printf("send_event: %2.2x\n", event->send_event);
		printf("time  : %8.8x\n", event->time);
		printf("state : %8.8x\n", event->state);
		printf("keyval: %8.8x\n", event->keyval);
		printf("length: %8.8x\n", event->length);
		printf("string: %s\n", event->string);
		printf("(hex) : %2.2x\n", event->string[0]);
		printf("cmd   : %s\n", cmd);
		printf("(hex) : %8.8x\n", cmd[0]);
#endif

		if ( !commandLine ) {
			ret = processKeyboardCommand( cmd );
		}
		else {
			ret = processCommandLine( cmd );
			cmd[ 0 ] = 0;
		}
	}
	return ret;
}

/** Process a menu command.

  	\param command	command to be processed
*/

gboolean PageTrim::processMenuCommand( char *command ) {
	strcpy( cmd, command );
	strcpy( lastcmd, command );
	gboolean ret = processKeyboardCommand( cmd );
	return ret;
}

/** Internal method for handling a complete keyboard scene.

  	\param cmd		command to be processed;
*/

gboolean PageTrim::processKeyboardCommand( char *cmd ) {
	int	end;
	GtkWidget *video_command = lookup_widget(common->getWidget(), "command");
	int count = 1;
	char real[ 256 ] = "";

	switch( sscanf( cmd, "%d%s", &count, real ) )
	{
		case 1:
			// Numeric value only - return immediately if the cmd is not "0"
			if ( strcmp( cmd, "0" ) ) {
				common->keyboardFeedback(cmd, "");
				return FALSE;
			}
			break;
		case 0:
			sscanf( cmd, "%s", real );
			count = 1;
			break;
	}

	if ( strcmp( cmd, "." ) && strcmp( cmd, ":" ) )
		strcpy( lastcmd, cmd );

	if ( !strcmp( cmd, ":" ) ) 
	{
		gtk_entry_set_editable( GTK_ENTRY( video_command ), TRUE );
		gtk_widget_grab_focus( video_command );
		gtk_entry_set_text( GTK_ENTRY( video_command ), "" );
		g_nav_ctl.escaped = TRUE;
		cmd[0] = 0;
	}
	/* Navigation */

	/* play, pause */
	
	else if (strcmp(cmd, " ") == 0) 
	{
		if ( g_nav_ctl.active == FALSE ) {
			common->keyboardFeedback(cmd, "Play");
			common->videoPlay( );
		} else {
			common->keyboardFeedback(cmd, "Pause");
			common->videoPause( );
		}
		cmd[0] = 0;
	}
	
	/* advance one frame */

	else if ( strcmp(real, "l") == 0 ) 
	{
		common->keyboardFeedback(cmd, "Move forward" );
		movedToFrame( pos + count);
		cmd[0] = 0;
	}

	/* backspace one frame */

	else if ( strcmp(real, "h") == 0 ) 
	{
   		common->keyboardFeedback(cmd, "Move backward");
		movedToFrame( pos - count );
		cmd[0] = 0;
	}

	/* advance one second */

	else if ( strcmp(real, "w") == 0 || strcmp(real, "W") == 0 ||
		  	strcmp(real, "e") == 0 || strcmp(real, "E") == 0 ) 
	{
		common->keyboardFeedback(cmd, "Move forward second");
		movedToFrame( pos + count * _getOneSecond() );
		cmd[0] = 0;
	}

	/* backspace one second */

	else if ((strcmp(real, "b") == 0) || (strcmp(real, "B") == 0)) 
	{
		common->keyboardFeedback(cmd, "Move backwards one second");
		moveToFrame( pos - count * _getOneSecond() );
		cmd[0] = 0;
	}

	/* start of scene */

	else if ((strcmp(cmd, "0") == 0) || (strcmp(real, "^") == 0)) 
	{
		common->videoStartOfScene( );
		common->keyboardFeedback(cmd, "Move to start of scene");
		cmd[0] = 0;
	}

	/* end of scene */

	else if (strcmp(real, "$") == 0) 
	{
		common->videoEndOfScene( );
		common->keyboardFeedback(cmd, "Move to end of scene");
		cmd[0] = 0;
	}

	/* start of next scene */

	else if ((strcmp(real, "j") == 0) || strcmp(real, "+") == 0 ) 
	{
		selectScene( getSceneIndex() +1);
		common->keyboardFeedback(cmd, "Move to start of next scene");
		cmd[0] = 0;
	}

	/* start of previous scene */

	else if ((strcmp(real, "k") == 0) || (strcmp(real, "-") == 0)) 
	{
		selectScene( getSceneIndex() -1);
		common->keyboardFeedback(cmd, "Move to start of previous scene");
		cmd[0] = 0;
	}

	/* first frame */

	else if (strcmp(cmd, "gg") == 0) 
	{
		common->videoStartOfMovie( );
		common->keyboardFeedback(cmd, "Move to first frame");
		cmd[0] = 0;
	}

	/* last frame */

	else if (strcmp(cmd, "G") == 0) 
	{
		common->videoEndOfMovie( );
		common->keyboardFeedback(cmd, "Move to last frame");
		cmd[0] = 0;
	}

	/* delete from current frame up to end of scene */

	else if (strcmp(cmd, "d$") == 0) 
	{
		out = pos;
		gtk_adjustment_set_value(trim_adj[TRIM_ADJ_OUT], (gfloat) out);
		common->keyboardFeedback(cmd, "Cut to end of scene");
		cmd[0] = 0;
	}

	/* delete from start of scene just before current frame */

	else if ((strcmp(cmd, "d0") == 0) || strcmp(cmd, "d^") == 0) 
	{
		in = pos;
		gtk_adjustment_set_value(trim_adj[TRIM_ADJ_IN], (gfloat) in);
		common->keyboardFeedback(cmd, "Cut from start of scene");
		cmd[0] = 0;
	}

	/* Switch to capture mode */

	else if ( strcmp( cmd, "i" ) == 0 ) 
	{
		common->keyboardFeedback(cmd, "Capture");
		FileTracker::GetInstance().SetMode( CAPTURE_FRAME_INSERT );
		common->changePageRequest( PAGE_CAPTURE );
		cmd[0] = 0;
	}

	else if ( strcmp( cmd, "a" ) == 0 ) 
	{
		common->keyboardFeedback(cmd, "Capture");
		end = common->getPlayList()->FindEndOfScene(common->g_currentFrame );
		common->moveToFrame( end );
		FileTracker::GetInstance().SetMode( CAPTURE_FRAME_APPEND );
		common->changePageRequest( PAGE_CAPTURE );
		cmd[0] = 0;
	}

	else if ( strcmp( cmd, "A" ) == 0 ) 
	{
		common->keyboardFeedback(cmd, "Capture");
		end = common->getPlayList()->GetNumFrames();
		common->moveToFrame( end );
		FileTracker::GetInstance().SetMode( CAPTURE_MOVIE_APPEND );
		common->changePageRequest( PAGE_CAPTURE );
		cmd[0] = 0;
	}

	else if ( strcmp( cmd, "u" ) == 0 ) 
	{
		common->keyboardFeedback(cmd, "Undo");
		resetInPoint();
		resetOutPoint();
		movedToFrame(pos);
		cmd[0] = 0;
	}
	
	else if ( strcmp( cmd, "\\" ) == 0 )
	{
		common->keyboardFeedback( cmd, "Toggle Looping" );
		gtk_toggle_button_set_active( loop_toggle, ! gtk_toggle_button_get_active(loop_toggle) );
		cmd[0] = 0;
	}

	else 
	{
		// Check for invalid commands
		if ( strlen( real ) > 3 )
			cmd[ 0 ] = 0;
		else if ( strchr( "dgy ", real[ strlen( real ) - 1 ] ) == NULL ) 
			cmd[ 0 ] = 0;

		common->keyboardFeedback(cmd, "");

	}

	return FALSE;
}

/** Process command line.

  	\param command	command to be processed.
*/

gboolean PageTrim::processCommandLine( char *command ) {

	if ( !strcmp( command, "." ) )
		strcpy( command, lastcmd );

	/* switch to export mode */

	else if (strcmp(command, ":W") == 0) {
		common->keyboardFeedback(command, "Export");
		common->changePageRequest( PAGE_EXPORT );
	}

	/* write PlayList */

	else if (strcmp(command, ":w") == 0) {
		common->keyboardFeedback(command, "Write playlist");
		common->savePlayList( );
	}

	/* quit */

	else if (strcmp(command, ":q") == 0) {
		common->keyboardFeedback(command, "quit");
		kinoDeactivate();
	} 
	
	/* goto a frame */
	else if ( strncmp( command, ":", 1 ) == 0 ) {
		int val = 0;
		char t[ 132 ] = "";
		if ( sscanf( command + 1, "%d", &val ) == 1 ) {
			common->moveToFrame( val - 1 );
			sprintf( t, "Move to frame %d", val );
			common->keyboardFeedback(command, t );
		}
	}

	/* cleanup the command entry */

	GtkWidget *video_command = lookup_widget(common->getWidget(), "command");
	g_nav_ctl.escaped = FALSE;
	gtk_entry_set_text( GTK_ENTRY(video_command), "");
	gtk_entry_set_editable( GTK_ENTRY(video_command), FALSE );
	gtk_widget_grab_focus( GTK_WIDGET(frameArea) );

	// Last command is now
	strcpy( lastcmd, command );

	return FALSE;
}


void PageTrim::windowChanged() {
	if ( max == 0 )
		common->clearPreview( frameArea );
	else
		showFrame( pos, TRUE );
}

void PageTrim::windowMoved() {
	if ( max == 0 )
		common->clearPreview( frameArea );
	else
		showFrame( pos, TRUE );
}


void PageTrim::showFrameInfo( int i ) {
	char msg[256];
	TimeCode tc;
	Frame &frame = *( GetFramePool( )->GetFrame( ) );
	if (i < 0) {
		sprintf(msg, "0 / 0\n--\n00:00:00:00"); 
	} else {
		playlist.GetFrame(i, frame);
		frame.GetRecordingDate();
		frame.GetTimeCode(tc);
		sprintf(msg, "%d / %d\n%s\n%2.2d:%2.2d:%2.2d:%2.2d", 
				i + 1, max+1,
				frame.GetRecordingDate().c_str(),
				tc.hour, tc.min, tc.sec, tc.frame );
		FileHandler *media;
		playlist.GetMediaObject( i, &media );
		common->showFrameMoreInfo( frame, media );
	}

	gtk_label_set_text( positionLabel, msg);
	GetFramePool( )->DoneWithFrame( &frame );
}


extern "C" {

	// Share the frames extracted from the audio and video thread - this
	// provides a smoother playback and is less intensive on the CPU. The
	// code assumes the video will never fall more than PLAYBACK_FRAMES behind 
	// the audio.
	//
	// 'position' stores last obtained index in the 'frameContent' array and the 
	// corresponding frame number is stored in the corresponding entry in the
	// frameNumber array.
	//
	// The audio thread is responsible for write access to these.

	static int frameNumber[ PLAYBACK_FRAMES ];
	static int pending = 0;
	static int head = -1;
	static int tail = -1;
	static bool showing = false;
	static bool playing = false;

	static void resetThreads( )
	{
		showing = false;
		playing = false;
		pending = 0;
		head = 0;
		tail = 0;
	}

	/** This function carries out the read ahead and is responsible for obtaining
		the each frame. audioThread monitors the queue (by checking mainly on pending)
		and plays the audio of each frame. videoThread picks up the last tail and
		displays it.

		Note that only half of the queue is filled here - this provides the video thread
		some time to safely use a frame without worrying about it being overwritten.
	*/

	static void *readThread( void *info ) {
		struct navigate_control *ctl = (struct navigate_control *)info;
		gint newFrame = 0;
		gint lastFrame = common->getPageTrim()->getPosition() - 1;
		gint totalFrames = common->getPageTrim()->getTotalFrames();
		gint countFrames = 0;
		static Preferences &prefs = Preferences::getInstance();
		int time_per_frame = 1000000 / _getOneSecond( );

		// cout << ">>> Starting read thread " << endl;

		// Get start of time
		struct timeval start;
		struct timeval end;
		gettimeofday( &start, NULL );

		// Loop while active
		while( ctl->active ) {

			// Calculate time for next frame
			start.tv_usec += time_per_frame;
			if ( start.tv_usec >= 1000000 )
			{
				start.tv_usec -= 1000000;
				start.tv_sec ++;
			}

			// Determine the frame to render
			newFrame = lastFrame + ctl->step;

			// determine new frame based upon jogshuttle rate
			if (ctl->step == 0) {
		  		ctl->subframe++;
		  		if (ctl->rate < 0) {
					if (ctl->subframe >= -ctl->rate) {
			  			newFrame --;
			  			ctl->subframe = 0;
					}
		  		} else {
					if (ctl->subframe >= ctl->rate) {
			  			newFrame ++;
			  			ctl->subframe = 0;
					}
		  		}
			}
	
			// Check the bounds and adjust as necessary
			if ( newFrame < 0 )
				newFrame = totalFrames-1;
			else if ( newFrame > common->getPageTrim()->getOutPoint() && 
					gtk_toggle_button_get_active(loop_toggle) == TRUE )
				newFrame = common->getPageTrim()->getInPoint();
			else if ( newFrame < common->getPageTrim()->getInPoint() && 
					gtk_toggle_button_get_active(loop_toggle) == TRUE )
				newFrame = common->getPageTrim()->getOutPoint();
			else if ( newFrame >= totalFrames )
				newFrame = 0;

			// Determine which locations in frameNumber and frameContent we need to use
			if ( pending >= PLAYBACK_FRAMES / 2 )
			{
				// Queue is full.. do nothing
				playing = true;
				struct timespec t = { 0, 0 };
				nanosleep( &t, NULL );
			}
			else if ( pending < PLAYBACK_FRAMES ) 
			{
				pthread_mutex_lock( &threadlock );
				int new_head = ( head + 1 ) % PLAYBACK_FRAMES;
				//cout << "captured pending = " << pending << " new_head = " << new_head << " tail = " << tail << endl;
				pthread_mutex_unlock( &threadlock );

				frameNumber[ new_head ] = newFrame;
				common->getPageTrim()->getPlayList().GetFrame( frameNumber[ new_head ], *frameContent[ new_head ] );

				pthread_mutex_lock( &threadlock );
				pending ++;
				head = new_head;
				pthread_mutex_unlock( &threadlock );

				if ( dv1394 != NULL )
					dv1394->SendFrame( *frameContent[tail] );
				
				// Inform the audio thread that it's got something to work with
				playing = true;

				// Render all frames if requested
				if ( ( ! prefs.dropFrame || ! prefs.enableAudio ) && ctl->active ) {

					// Change the queue position
					tail = ( tail + 1 ) % PLAYBACK_FRAMES;
					pending --;

					// Set the current position
					common->getPageTrim()->setPosition( frameNumber[ tail ] );

					// If audio is enabled and some other conditions
					if ( prefs.enableAudio && ( frameNumber[ tail ] != lastFrame  || ctl->rate > 1 ) )
						common->getPageTrim()->getFrameDisplayer()->PutSound( *frameContent[ tail ] );

					// Show the corresponding image
					gdk_threads_enter();
					common->getPageTrim()->showFrame( frameNumber[ tail ], *frameContent[ tail ] );
					gdk_flush();
					gdk_threads_leave();

					// Determine how far from the next frame we really are
					gettimeofday( &end, NULL );
					int difference = ( ( start.tv_sec * 1000000 + start.tv_usec ) - ( end.tv_sec * 1000000 + end.tv_usec ) );

					if ( difference > 2000 && difference < time_per_frame )
					{
						// Sleep for half the remaining time when audio is disabled
						if ( ! prefs.enableAudio )
						{
							struct timespec t = { 0, difference * 500 };
							nanosleep( &t, NULL );
						}
					}
					else if ( difference < 0 || difference >= time_per_frame )
					{
						gettimeofday( &start, NULL );
					}
				}

				// update the lastFrame and lastPos variables
				lastFrame = newFrame;

				// Incrment the frame count
				countFrames ++;
			}
		}

		// cout << ">>> Ending reading thread" << endl;

		return NULL;
	}

	static void *audioThread( void *info ) {

		static Preferences &prefs = Preferences::getInstance();
		struct navigate_control *ctl = (struct navigate_control *)info;
		gint lastFrame = common->getPageTrim()->getPosition() - 1;
		int time_per_frame = 1000000 / _getOneSecond( );

		// cout << ">>> Starting audio thread" << endl;

		while( ctl->active ) {

			// Only do this when drop frames is active (otherwise the audio thread takes care of it)
			if ( prefs.dropFrame && prefs.enableAudio ) {

				// Obtain the position of the last released frame
				pthread_mutex_lock( &threadlock );
				int new_tail = ( tail + 1 ) % PLAYBACK_FRAMES;
				//cout << "playing  pending = " << pending << " head = " << head << " new_tail = " << new_tail << endl;
				pthread_mutex_unlock( &threadlock );

				// As long as its different to the previous one, then display it
				if ( pending > 0 && playing ) {

					// Move the main frame position to our newly derived location
					common->getPageTrim()->setPosition( frameNumber[ tail ] );

					// Play the audio
					if ( prefs.enableAudio && ( frameNumber[ tail ] != lastFrame || ctl->rate > 1 ) )
					{
						common->getPageTrim()->getFrameDisplayer()->PutSound( *frameContent[ tail ] );
						if ( pending < 10 )
						{
							// Give the reader more time 
							struct timespec t = { 0, 0 };
							nanosleep( &t, NULL );
						}
					}
					else if ( ctl->step == 0 && !prefs.audioScrub )
					{
						struct timespec t = { 0, time_per_frame * 1000 };
						nanosleep( &t, NULL );
					}

					// Last encountered audio frame
					lastFrame = frameNumber[ tail ];

					// Move the tail
					pthread_mutex_lock( &threadlock );
					pending --;
					tail = new_tail;
					pthread_mutex_unlock( &threadlock );

					// Start showing frames from this point onwards
					showing = true;

				}
				else
				{
					struct timespec t = { 0, 0 };
					nanosleep( &t, NULL );
				}
			}
			else {
				// It is safe to cancel the audio thread here since a change to prefs necessitates a restart.
				// cout << ">>> Audio Thread not needed" << endl;
				return NULL;
			}
		}

		// cout << ">>> Ending video thread" << endl;

		return NULL;
	}

	static void *videoThread( void *info ) {

		static Preferences &prefs = Preferences::getInstance();
		struct navigate_control *ctl = (struct navigate_control *)info;
		int lastFrame = common->getPageTrim()->getPosition() - 1;

#ifdef PLAY_WITH_STATS
		// Statistical analysis variables
		int dropped = 0;
		int count = 0;
#endif

		// cout << ">>> Starting video thread" << endl;

		while( ctl->active ) {

			// Only do this when drop frames is active (otherwise the audio thread takes care of it)
			if ( prefs.dropFrame && prefs.enableAudio ) {

				if ( showing && pending > 0 ) {

					// Store the current tail
					int my_tail = tail;

					// As long as its different to the previous one, then display it
					if ( frameNumber[ my_tail ] != lastFrame || ctl->rate > 1 )
					{
#ifdef PLAY_WITH_STATS
						// Statistical analysis of frame playback (useful for limited kinds of tests)
						if ( ctl->step == 1 ) {
							int skipped = frameNumber[ my_tail ] - lastFrame - 1;
							count += skipped + 1;
							if ( skipped != 0 && lastFrame != 0 && count != 0 ) {
								dropped += skipped;
								cout << ">>>> Dropped " << dropped << " in " << count << " frames - " << (double)( ( (double)dropped / (double)count ) * 100.0 ) << "%" << endl;
							}
							lastFrame = frameNumber[ my_tail ];
						}
#endif

						// Show this frame
						gdk_threads_enter();
						if ( ctl->active ) {
							common->getPageTrim()->showFrame( frameNumber[ my_tail ], *frameContent[ my_tail ] );
							gdk_flush();
						}
						gdk_threads_leave();
					}

					// Always release after showing a frame (hmm...)
					struct timespec t = { 0, 0 };
					nanosleep( &t, NULL );
				}
			}
			else {
				// It is safe to cancel the video thread here since a change to prefs necessitates a restart.
				// cout << ">>> Video Thread not needed" << endl;
				return NULL;
			}
		}

#ifdef PLAY_WITH_STATS
		cout << ">>>> Video stopped: Dropped " << dropped << " in " << count << " frames - " << (double)( ( (double)dropped / (double)count ) * 100.0 ) << "%" << endl;
#endif

		// cout << ">>> Ending video thread" << endl;

		return NULL;
	}

	static int _getOneSecond(void)
	{
		Frame &frame = *( GetFramePool()->GetFrame( ) );
		common->getPlayList()->GetFrame(common->g_currentFrame, frame);
		int value = (frame.IsPAL() ? 25 : 30);
		GetFramePool( )->DoneWithFrame( &frame );
		return value;
	}

	
	gboolean
	on_start_spin_focus_in_event           (GtkWidget       *widget,
											GdkEventFocus   *event,
											gpointer         user_data)
	{
		g_nav_ctl.escaped = TRUE;
		return FALSE;
	}
	
	
	gboolean
	on_start_spin_focus_out_event          (GtkWidget       *widget,
											GdkEventFocus   *event,
											gpointer         user_data)
	{
		g_nav_ctl.escaped = FALSE;
		return FALSE;
	}
	
	
	gboolean
	on_end_spin_focus_in_event             (GtkWidget       *widget,
											GdkEventFocus   *event,
											gpointer         user_data)
	{
		g_nav_ctl.escaped = TRUE;
		return FALSE;
	}
	
	
	gboolean
	on_end_spin_focus_out_event            (GtkWidget       *widget,
											GdkEventFocus   *event,
											gpointer         user_data)
	{
		g_nav_ctl.escaped = FALSE;
		return FALSE;
	}
	
}

