/*
 * Copyright (C) 2001 Dan Dennedy <dan@dennedy.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "preferences.h"
#include "kino_common.h"

// #include <string.h>
#include <libgnome/gnome-config.h>

#include <string>
using std::string;
using std::make_pair;
/** The global prefs object.
    
    It is accessible to other classes via the getInstance method.
*/

Preferences *Preferences::instance = NULL;

/** Singleton getInstance method. Standard pattern followed, non threadsafe
	but not required (at the moment) since it gets called before any threading
	is activated. Keep an eye on it though.

	\returns the single instance of Preferences.
*/

Preferences &Preferences::getInstance() {
	if ( instance == NULL )
		instance = new Preferences;
	return *instance;
}

/** The constructor loads the user preferences.

    Since this uses gnome-config, preferences are typically
    loaded from ~/.gnome/kino.
*/
Preferences::Preferences() {
    gchar key[256];
    gchar *tempString;

    gnome_config_push_prefix("/kino/general/");

    tempString = gnome_config_get_string("file=capture");
    if (tempString) {
        strncpy( file, tempString, 511);
        g_free(tempString);
    }
    defaultNormalisation = gnome_config_get_int("defaultNormalisation=0");
    defaultAudio = gnome_config_get_int("defaultAudio=0");
    defaultAspect = gnome_config_get_int("defaultAspect=0");
    g_snprintf( key, 255, "fileFormat=%d", AVI_DV1_FORMAT);
    fileFormat = gnome_config_get_int(key);
    autoSplit = gnome_config_get_bool("autoSplit=true");
    timeStamp = gnome_config_get_bool("timeStamp=false");
    frames = gnome_config_get_int("frames=7000");
    every = gnome_config_get_int("every=1");
    interface = gnome_config_get_int("interface=0");
    channel = gnome_config_get_int("channel=63");
	phyID = -1;
	tempString = gnome_config_get_string("avcGUID=");
    if (tempString) {
        strncpy( avcGUID, tempString, 64);
        g_free(tempString);
    }
    g_snprintf( key, 255, "displayMode=%d", DISPLAY_XV);
    displayMode = gnome_config_get_int(key);
    displayQuality = gnome_config_get_float("displayQuality=2");
    displayFixed = gnome_config_get_bool("displayFixed=false");
    enableAudio = gnome_config_get_bool("enableAudio=true");
    cip_n = gnome_config_get_int("cip_increment=0");
    cip_d = gnome_config_get_int("cip_threshold=0");
    syt_offset = gnome_config_get_int("syt_offset=19000");
    preview_capture = gnome_config_get_bool("preview_capture=false");
    dropFrame = gnome_config_get_bool("dropFrame=true");
    tempString = gnome_config_get_string("audioDevice=/dev/dsp");
    if (tempString) {
        strncpy( audioDevice, tempString, 511);
        g_free(tempString);
    }

    /// Jogshuttle stuff
    enableJogShuttle = gnome_config_get_bool("enableJogShuttle=false");
    tempString = gnome_config_get_string("jogShuttleDevice=/dev/input/event0");
    if (tempString) {
        strncpy( jogShuttleDevice, tempString, 511);
        g_free(tempString);
    }
    _isGrabbing = false;
    /* Fill array of actions */
    /* Element 0 is empty */
    int count = 0; /** Ease moving code around */
    /* "<none>" should always go first */
    _JogShuttleActions.push_back( 
	    JogShuttleActions( count++, "<none>",
			       "", 
			       "" ) );

    _JogShuttleActions.push_back(
	    JogShuttleActions( count++, 
			       "Toggle play", 
			       "Toggle between play and pause",
			       " " ) );

    /* TODO
    _JogShuttleActions.push_back(
	    JogShuttleActions( count++, 
			       "Stop",
			       "Stop playback", 
			       "Esc" ) );
    */
    
    /* --- NAVIGATION --- */

    _JogShuttleActions.push_back( 
	    JogShuttleActions( count++, "Forward one frame",
			       "Move one frame forward", 
			       "l" ) );

    _JogShuttleActions.push_back( 
	    JogShuttleActions( count++, "Back one frame",
			       "Move one frame backward", 
			       "h" ) );

    _JogShuttleActions.push_back(
	    JogShuttleActions( count++, "Forward one second", 
			       "Move one second forward",
			       "w" ) );

    _JogShuttleActions.push_back(
	    JogShuttleActions( count++, "Back one second", 
			       "Move one second backward", 
			       "b") );
    
    _JogShuttleActions.push_back(
	    JogShuttleActions( count++, "Start of scene", 
			       "Move to start of current scene",
			       "0" ) );

    _JogShuttleActions.push_back(
	    JogShuttleActions( count++, "End of scene", 
			       "Move to end of current scene",
			       "$" ) );

    _JogShuttleActions.push_back(
	    JogShuttleActions( count++, "Previous scene", 
			       "Move to start of previous scene",
			       "k" ) );

    _JogShuttleActions.push_back(
	    JogShuttleActions( count++, "Next scene", 
			       "Move to start of next scene",
			       "j" ) );

    _JogShuttleActions.push_back(
	    JogShuttleActions( count++, "First frame", 
			       "Move to beginning of movie",
			       "gg" ) );

    _JogShuttleActions.push_back(
	    JogShuttleActions( count++, "Last frame", 
			       "Move to end of movie",
			       "G" ) );
    
    _JogShuttleActions.push_back(
	    JogShuttleActions( count++, "Forward 5 scenes", 
			       "Move forward 5 scenes",
			       "5$" ) );

    _JogShuttleActions.push_back(
	    JogShuttleActions( count++, "Back 5 scene", 
			       "Move backward 5 scenes",
			       "5^" ) );


    /* --- CUT --- */
    
    _JogShuttleActions.push_back( 
	    JogShuttleActions( count++, "Cut current frame",
			       "Cut current frame", 
			       "x" ) );

    _JogShuttleActions.push_back( 
	    JogShuttleActions( count++, "Cut a second",
			       "Cut a second", 
			       "dw" ) );

    _JogShuttleActions.push_back( 
	    JogShuttleActions( count++, "Cut current scene",
			       "Cut current scene", 
			       "dd" ) );

    _JogShuttleActions.push_back( 
	    JogShuttleActions( count++, "Cut to end of current scene",
			       "Cut to end of current scene", 
			       "d$" ) );

    _JogShuttleActions.push_back( 
	    JogShuttleActions( count++, "Cut to end of movie",
			       "Cut to end of movie", 
			       "dG" ) );

    _JogShuttleActions.push_back( 
	    JogShuttleActions( count++, "Cut from start of current scene",
			       "Cut from start of current scene", 
			       "d0" ) );

    _JogShuttleActions.push_back( 
	    JogShuttleActions( count++, "Cut from start of movie",
			       "Cut from start of movie", 
			       "dgg" ) );

    /* --- COPY --- */

    _JogShuttleActions.push_back( 
	    JogShuttleActions( count++, "Copy current scene",
			       "Copy current scene", 
			       "yy" ) );

    _JogShuttleActions.push_back( 
	    JogShuttleActions( count++, "Copy to end of scene",
			       "Copy to end of scene", 
			       "y$" ) );

    _JogShuttleActions.push_back( 
	    JogShuttleActions( count++, "Copy from start of scene",
			       "Cut from start of scene", 
			       "y0" ) );

    /* --- PASTE --- */

    _JogShuttleActions.push_back( 
	    JogShuttleActions( count++, "Paste after current frame",
			       "Paste after current frame", 
			       "p" ) );

    _JogShuttleActions.push_back( 
	    JogShuttleActions( count++, "Paste before current frame",
			       "Paste before current frame", 
			       "P" ) );

    /* --- MODE SWITCHING --- */
    
    _JogShuttleActions.push_back( 
	    JogShuttleActions( count++, "Switch to capture mode",
			       "Switch to the capture interface", 
			       "i" ) );

    _JogShuttleActions.push_back( 
	    JogShuttleActions( count++, "Switch to trim mode",
			       "Switch to trim mode", 
			       "t" ) );
    /* --- GENERAL --- */

    _JogShuttleActions.push_back( 
	    JogShuttleActions( count++, "Repeat last command",
			       "Repeat last command", 
			       "." ) );

    _JogShuttleActions.push_back( 
	    JogShuttleActions( count++, "Split scene",
			       "Split scene before current frame", 
			       ":split" ) );
    
    _JogShuttleActions.push_back( 
	    JogShuttleActions( count++, "Join scenes",
			       "Join this scene with the following", 
			       "J" ) );

    _JogShuttleActions.push_back( 
	    JogShuttleActions( count++, "Undo",
			       "Undo last command", 
			       "u" ) );

    _JogShuttleActions.push_back( 
	    JogShuttleActions( count++, "Redo",
			       "Redo last command", 
			       "ctrl+r" ) );

    // TODO: Add more actions? 


    /* Actually get the preferences */
    count = gnome_config_get_int( "jogShuttleNumMappings=0" );
    if ( 0 != count ) 
    {
	    for ( int i = 0; i < count ; i++ )
	    {
		    char key[ 512 ];
		    sprintf( key, "jogShuttleMappingFirstButton%i=0", i );
		    unsigned int first_code 
			    = gnome_config_get_int( key ); 
		    sprintf( key, "jogShuttleMappingSecondButton%i=0", i );
		    unsigned int second_code 
			    = gnome_config_get_int( key ); 
		    sprintf( key, "jogShuttleMappingAction%i=", i );
		    // g_print ( " Key = %s\n", key );
		    tempString = 
			    gnome_config_get_string( key );

		    // TODO: Free
		    // g_print( "tmpstring Read action %s\n", tempString );
		    string action = string( tempString );
		    // g_print( "Read action %s\n", action.c_str() );
		    /* Insert */
		    _JogShuttleMappings[ make_pair ( first_code, second_code ) ]
			    = JogShuttleMapping( action );
	    }
    }
    else
    {
	    // Set up defaults manuall, if 0 count
	    // These are my own default mappings
	    _JogShuttleMappings[ make_pair ( 0x100, 0x0 ) ] 
		    = JogShuttleMapping( "d0" );
	    _JogShuttleMappings[ make_pair ( 0x101, 0x0 ) ] 
		    = JogShuttleMapping( "d$" );
	    _JogShuttleMappings[ make_pair ( 0x102, 0x0 ) ] 
		    = JogShuttleMapping( "h" );
	    _JogShuttleMappings[ make_pair ( 0x103, 0x0 ) ] 
		    = JogShuttleMapping( "l" );
	    _JogShuttleMappings[ make_pair ( 0x104, 0x0 ) ] 
		    = JogShuttleMapping( "b" );
	    _JogShuttleMappings[ make_pair ( 0x105, 0x0 ) ] 
		    = JogShuttleMapping( "w" );
	    _JogShuttleMappings[ make_pair ( 0x106, 0x0 ) ] 
		    = JogShuttleMapping( " " );
	    _JogShuttleMappings[ make_pair ( 0x107, 0x0 ) ] 
		    = JogShuttleMapping( "k" );
	    _JogShuttleMappings[ make_pair ( 0x108, 0x0 ) ] 
		    = JogShuttleMapping( "j" );
	    _JogShuttleMappings[ make_pair ( 0x109, 0x0 ) ] 
		    = JogShuttleMapping( "gg" );
	    _JogShuttleMappings[ make_pair ( 0x10a, 0x0 ) ] 
		    = JogShuttleMapping( "G" );
    }


    
    
	enableV4L = gnome_config_get_bool( "enableV4L=false" );
	disableKeyRepeat = gnome_config_get_bool( "disableKeyRepeat=false");
	audioRendering = gnome_config_get_int( "audioRendering=0" );
	previewRendering = gnome_config_get_int( "previewRendering=1" );
	enableTimecode = gnome_config_get_bool( "enableTimecode=true" );
	maxUndos = gnome_config_get_int( "maxUndos=50" );
	//tempString = gnome_config_get_string("video1394Device=/dev/video1394");
    //if (tempString) {
    //    strncpy( video1394Device, tempString, 511);
    //    g_free(tempString);
    //}
	dvCaptureBuffers = gnome_config_get_int( "dvCaptureBuffers=50" );
	dvExportBuffers = gnome_config_get_int( "dvExportBuffers=10" );
	dvDecoderAddNTSCSetup = gnome_config_get_bool( "dvDecoderAddNTSCSetup=false" );
	dvDecoderClampLuma = gnome_config_get_bool( "dvDecoderClampLuma=false" );
	dvDecoderClampChroma = gnome_config_get_bool( "dvDecoderClampChroma=false" );
	maxFileSize = gnome_config_get_int( "maxFileSize=2000" );
	audioScrub = gnome_config_get_bool( "audioScrub=false" );
  	tempString = gnome_config_get_string( "v4lVideoDevice=/dev/video0" );
    if (tempString) {
        strncpy( v4lVideoDevice, tempString, 511);
        g_free(tempString);
    }
	tempString = gnome_config_get_string( "v4lAudioDevice=/dev/dsp" );
    if (tempString) {
        strncpy( v4lAudioDevice, tempString, 511);
        g_free(tempString);
    }
	tempString = gnome_config_get_string( "v4lInput=PAL" );
    if (tempString) {
        strncpy( v4lInput, tempString, 31);
        g_free(tempString);
    }
	tempString = gnome_config_get_string( "v4lAudio=32000" );
    if (tempString) {
        strncpy( v4lAudio, tempString, 31);
        g_free(tempString);
    }
    isOpenDML = gnome_config_get_bool( "isOpenDML=false" );
	tempString = gnome_config_get_string( "defaultDirectory=~/" );
    if (tempString) {
        strncpy( defaultDirectory, tempString, 511);
        g_free(tempString);
    }
    deinterlacePreview = gnome_config_get_bool( "deinterlacePreview=false" );
    relativeSave = gnome_config_get_bool( "relativeSave=false" );

	dvCaptureDriver = gnome_config_get_int( "dvCaptureDriver=0" );
	tempString = gnome_config_get_string( "dvCaptureDevice=/dev/raw1394" );
    if (tempString) {
        strncpy( dvCaptureDevice, tempString, 511);
        g_free(tempString);
    }
	dv1394Preview = gnome_config_get_bool( "dv1394Preview=false" );
	dvExportDriver = gnome_config_get_int( "dvExportDriver=0" );
	tempString = gnome_config_get_string( "dvExportDevice=/dev/video1394" );
    if (tempString) {
        strncpy( dvExportDevice, tempString, 511);
        g_free(tempString);
    }
	avcPollIntervalMs = gnome_config_get_int( "avcPollIntervalMs=200" );
	dvExportPrerollSec = gnome_config_get_int( "dvExportPrerollSec=4" );
	dvTwoPassEncoder = gnome_config_get_bool( "dvTwoPassEncoder=false" );

    gnome_config_pop_prefix();
}

/** Save the user preferences.
*/

/// Save the user preferences.
void
Preferences::Save() {
    gnome_config_push_prefix("/kino/general/");

    gnome_config_set_int("defaultNormalisation", defaultNormalisation);
    gnome_config_set_int("defaultAudio", defaultAudio);
    gnome_config_set_int("defaultAspect", defaultAspect);
    gnome_config_set_string( "file", file);
    gnome_config_set_int("fileFormat", fileFormat);
    gnome_config_set_bool("autoSplit", autoSplit);
    gnome_config_set_bool("timeStamp", timeStamp);
    gnome_config_set_int("frames", frames);
    gnome_config_set_int("every", every);
    gnome_config_set_int("interface", interface);
    gnome_config_set_int("channel", channel);
	gnome_config_set_string("avcGUID", avcGUID);
    gnome_config_set_int("displayMode", displayMode);
    gnome_config_set_float("displayQuality", displayQuality);
    gnome_config_set_bool("displayFixed", displayFixed);
    gnome_config_set_bool("enableAudio", enableAudio);
    gnome_config_set_int("cip_increment", cip_n);
    gnome_config_set_int("cip_threshold", cip_d);
    gnome_config_set_int("syt_offset", syt_offset);
    gnome_config_set_bool("preview_capture", preview_capture);
    gnome_config_set_bool("dropFrame", dropFrame);
    gnome_config_set_string("audioDevice", audioDevice);

    /* JogShuttle Stuff */
    gnome_config_set_bool( "enableJogShuttle", enableJogShuttle );
    gnome_config_set_string( "jogShuttleDevice", jogShuttleDevice );
    gnome_config_set_int( "jogShuttleNumMappings", _JogShuttleMappings.size() );
    int count = 0;
    map<pair<unsigned short, unsigned short>, JogShuttleMapping>::iterator i;
    pair<unsigned short, unsigned short> map_key;
    for ( i = _JogShuttleMappings.begin(); 
	  i != _JogShuttleMappings.end() ; i++, count++ ) 
    {
	    map_key = ( *i ).first;
	    char key[ 512 ];
	    sprintf( key, "jogShuttleMappingFirstButton%i", count );
	    gnome_config_set_int( key, map_key.first ); 
	    sprintf( key, "jogShuttleMappingSecondButton%i", count );
	    gnome_config_set_int( key, map_key.second ); 
	    sprintf( key, "jogShuttleMappingAction%i", count );
	    gnome_config_set_string( key, ( *i ).second._action.c_str() ); 
    }

    gnome_config_set_bool("enableV4L", enableV4L);
    gnome_config_set_bool("disableKeyRepeat", disableKeyRepeat);
    gnome_config_set_int("audioRendering", audioRendering);
    gnome_config_set_int("previewRendering", previewRendering);
    gnome_config_set_bool("enableTimecode", enableTimecode);
    gnome_config_set_int("maxUndos", maxUndos);
	gnome_config_set_string("video1394Device", video1394Device);
	gnome_config_set_int("dvCaptureBuffers", dvCaptureBuffers);
	gnome_config_set_int("dvExportBuffers", dvExportBuffers);
	gnome_config_set_bool("dvDecoderAddNTSCSetup", dvDecoderAddNTSCSetup);
	gnome_config_set_bool("dvDecoderClampLuma", dvDecoderClampLuma);
	gnome_config_set_bool("dvDecoderClampChroma", dvDecoderClampChroma);
	gnome_config_set_int("maxFileSize", maxFileSize);
	gnome_config_set_bool("audioScrub", audioScrub);
	gnome_config_set_string("v4lVideoDevice", v4lVideoDevice);
	gnome_config_set_string("v4lAudioDevice", v4lAudioDevice);
	gnome_config_set_string("v4lInput", v4lInput);
	gnome_config_set_string("v4lAudio", v4lAudio);
    gnome_config_set_bool("isOpenDML", isOpenDML);
	gnome_config_set_string("defaultDirectory", defaultDirectory);
    gnome_config_set_bool("deinterlacePreview", deinterlacePreview);
    gnome_config_set_bool("relativeSave", relativeSave);
	
	gnome_config_set_int( "dvCaptureDriver", dvCaptureDriver );
	gnome_config_set_string( "dvCaptureDevice", dvCaptureDevice );
	gnome_config_set_bool( "dv1394Preview", dv1394Preview );
	gnome_config_set_int( "dvExportDriver", dvExportDriver );
	gnome_config_set_string( "dvExportDevice", dvExportDevice );
	gnome_config_set_int( "avcPollIntervalMs", avcPollIntervalMs );
	gnome_config_set_bool( "dvTwoPassEncoder", dvTwoPassEncoder );
	
    gnome_config_pop_prefix();

    gnome_config_sync();
}

/// Provide access to the audioDevice propery from C modules.
gchar *Preferences_getAudioDevice(void)
{
    return Preferences::getInstance().audioDevice;
}

